import logging

from .base_command import BaseCommand


class Settings(BaseCommand):
    """
    Portal Settings APIs

    :ivar cterasdk.core.settings.GlobalSettings global_settings: Object holding the Portal Global Settings APIs
    """

    def __init__(self, portal):
        super().__init__(portal)
        self.global_settings = GlobalSettings(self._core)


class GlobalSettings(BaseCommand):

    def get_timezone(self):
        """
        Get timezone
        """
        return self._core.api.get('/settings/timezone')

    def set_timezone(self, timezone):
        """
        Set timezone

        :param str timezone: Timezone
        """
        logging.getLogger().info('Updating timezone. %s', {'timezone': timezone})
        response = self._core.api.put('/settings/timezone', timezone)
        logging.getLogger().info('Updated timezone. %s', {'timezone': timezone})
        return response



import logging

from ..common import Object
from ..exceptions import CTERAException
from .base_command import BaseCommand


class Telnet(BaseCommand):
    """ Edge Filer Telnet configuration APIs """

    def enable(self, code):
        """ Enable Telnet """
        param = Object()
        param.code = code

        logging.getLogger().info("Enabling telnet access.")

        response = self._edge.api.execute("/config/device", "startTelnetd", param)
        if response == 'telnetd already running':
            logging.getLogger().info("Telnet access already enabled.")
        elif response != "OK":
            logging.getLogger().error("Failed enabling telnet access. %s", {'reason': response})
            raise CTERAException("Failed enabling telnet access", None, reason=response)
        else:
            logging.getLogger().info("Telnet access enabled.")

    def disable(self):
        """ Disable Telnet """
        logging.getLogger().info("Disabling telnet access.")
        self._edge.api.execute("/config/device", "stopTelnetd")
        logging.getLogger().info("Telnet access disabled.")



import logging
import functools


def update_current_tenant(execute_request):
    @functools.wraps(execute_request)
    def execute_and_update(self, tenant):
        logging.getLogger().debug('Updating current tenant. %s', {'tenant': tenant})
        return_value = execute_request(self, tenant)
        self.session().update_tenant(tenant)
        logging.getLogger().debug('Updated current tenant. %s', {'tenant': tenant})
        return return_value
    return execute_and_update



import logging
from .base_command import BaseCommand
from . import query
from ..exceptions import CTERAException


class RansomProtect(BaseCommand):
    """
    Ransomware Protect APIs
    """

    def get_configuration(self):
        """
        Get Ransom Protect Configuration

        :return cterasdk.common.object.Object: Configuration
        """
        return self._edge.api.get('/config/ransomProtect/')

    def enable(self):
        """Enable Ransom Protect service"""
        logging.getLogger().info('Enabling Ransom Protect.')
        self._edge.api.put('/config/ransomProtect/enabled', True)
        logging.getLogger().info('Ransom Protect enabled.')

    def disable(self):
        """Enable Ransom Protect service"""
        logging.getLogger().info('Disabling Ransom Protect.')
        self._edge.api.put('/config/ransomProtect/enabled', False)
        logging.getLogger().info('Ransom Protect disabled.')

    def is_disabled(self):
        """Check if Ransom Protect is disabled"""
        return self._edge.api.get('/config/ransomProtect/enabled') is not True

    def modify(self, block_users=None, detection_threshold=None, detection_interval=None):
        """
        Modify Ransom Protect Configuration.

        :param bool,optional block_users: Enable/Disable Block Users
        :param int,optional detection_threshold: Detection threshold (number of events)
        :param int,optional detection_interval: Detection interval (seconds)
        """
        param = self.get_configuration()
        if not param.enabled:
            raise CTERAException('Ransom Protect must be enabled to modify its configuration')
        if block_users is not None:
            param.shouldBlockUser = block_users
        if detection_threshold is not None:
            param.minimalNumOfFilesForPositiveDetection = detection_threshold
        if detection_interval is not None:
            param.detectionInterval = detection_interval
        self._edge.api.put('/config/ransomProtect/', param)

    def incidents(self):
        """
        List Ransomware Incidents

        :return list[cterasdk.common.object.Object]: List of incidents
        """
        return self._edge.api.execute('/proc/rpsrv', 'getListOfIncidents')

    def details(self, incident):
        """
        Retrieve Ransomware Incident Details

        :param int incident: Incident identifier, or an incident object
        """
        param = query.QueryParamBuilder()
        param.put('incidentId', incident if isinstance(incident, int) else incident.incident_id)
        return query.iterator(self._edge, '/proc/rpsrv', param.build(), 'getIncidentDetails')



import logging

from .enum import Role, RoleResolver
from .types import RoleSettings
from .base_command import BaseCommand


class Roles(BaseCommand):
    """
    Role Settings APIs
    """

    @staticmethod
    def types():
        """
        Get a list of roles
        """
        return [v for k, v in Role.__dict__.items() if not k.startswith('_')]

    @staticmethod
    def find(role):
        """
        Find Role
        """
        options = {k: v for k, v in RoleResolver.__dict__.items() if not k.startswith('_')}
        return options.get(role, None)

    def get(self, role):
        """
        Get Role

        :param str role: Role
        :returns: Role settings
        :rtype: cterasdk.core.types.RoleSettings
        """
        role = Roles.find(role)
        if role:
            return RoleSettings.from_server_object(self._core.api.get(f'/rolesSettings/{role}'))
        logging.getLogger().warning('Could not find role. %s', {'role': role})
        return None

    def modify(self, role, settings):
        """
        Modify Role

        :param cterasdk.core.enum.Role role: Role
        :param cterasdk.core.types.RoleSettings settings: Role Settings
        :returns: Updated role settings
        :rtype: cterasdk.core.types.RoleSettings
        """
        role = Roles.find(role)
        if role:
            logging.getLogger().info('Updating role settings. %s', {'role': role})
            response = self._core.api.put(f'/rolesSettings/{role}', settings.to_server_object())
            logging.getLogger().info('Role settings updated. %s', {'role': role})
            return RoleSettings.from_server_object(response)
        logging.getLogger().warning('Could not find role. %s', {'role': role})
        return None



import logging

from .enum import Mode
from ..exceptions import CTERAException
from .base_command import BaseCommand


class NFS(BaseCommand):
    """ Edge Filer NFS configuration """

    def get_configuration(self):
        """
        Get the current NFS configuration

        :return cterasdk.common.object.Object:
        """
        return self._edge.api.get('/config/fileservices/nfs')

    def enable(self):
        """ Enable NFS """
        self._set_mode(True)

    def disable(self):
        """ Disable NFS """
        self._set_mode(False)

    def is_disabled(self):
        """ Check if the NFS server is disabled """
        return self._edge.api.get('/config/fileservices/nfs/mode') == Mode.Disabled

    def _set_mode(self, enabled):
        logging.getLogger().info('%s NFS server.', ('Enabling' if enabled else 'Disabling'))
        self._edge.api.put('/config/fileservices/nfs/mode', Mode.Enabled if enabled else Mode.Disabled)
        logging.getLogger().info('NFS server %s.', ('enabled' if enabled else 'disabled'))

    def modify(
            self,
            async_write=None,
            aggregate_writes=None,
            mountd_port=None,
            statd_port=None,
            nfsv4_enabled=None,
            krb5_enabled=None,
            nfsd_host=None):
        """
        Modify the FTP Configuration. Parameters that are not passed will not be affected

        :param bool,optional async_write: If True, use asynchronous writes
        :param bool,optional aggregate_writes: If True, aggregate write requests
        :param int,optional mountd_port: Instruct mountd to bind to a specific port
        :param int,optional statd_port: Instruct statd to bind to a specific port
        :param bool,optional nfsv4_enabled: Enable NFSv4
        :param bool,optional krb5_enabled: Enable Kerberos. Note that NFS4V must be enabled to enable Kerberos
        :param str,optional nfsd_host: Instruct nfsd to bind to a specific network interface. Set to an empty string to clear
        """
        config = self.get_configuration()
        if config.mode != Mode.Enabled:
            raise CTERAException("NFS must be enabled in order to modify its configuration")
        if async_write is not None:
            setattr(config, 'async', Mode.Enabled if async_write else Mode.Disabled)
        if aggregate_writes is not None:
            config.aggregateWrites = Mode.Enabled if aggregate_writes else Mode.Disabled
        if mountd_port is not None:
            config.mountdPort = mountd_port
        if statd_port is not None:
            config.statdPort = statd_port
        if nfsv4_enabled is not None:
            config.nfsv4enabled = nfsv4_enabled
        if krb5_enabled is not None:
            if krb5_enabled and config.nfsv4enabled is not None and not config.nfsv4enabled:
                raise CTERAException("NFSv4 must be enabled in order to enable Kerberos")
            config.krb5 = krb5_enabled
        if nfsd_host is not None:
            config.nfsHost = nfsd_host
        self._edge.api.put('/config/fileservices/nfs', config)



import aiohttp

from ..common import Object
from ..convert import fromjsonstr, fromxmlstr
from ..exceptions import ClientResponseException


class ClientError(Object):

    def __init__(self, exception, message):
        self.request = Object()
        self.request.method = exception.request_info.method
        self.request.url = str(exception.request_info.real_url)
        self.response = Object()
        self.response.status = exception.status
        self.response.message = fromxmlstr(message) or fromjsonstr(message) or message


def accept(response, error_message):
    try:
        response.raise_for_status()
    except aiohttp.ClientResponseError as exception:
        error_object = ClientError(exception, error_message)
        raise ClientResponseException(error_object)
    return response



import logging

from ..common import Object
from .base_command import BaseCommand


class Licenses(BaseCommand):
    """
    Portal Licenses APIs
    """

    def all(self):
        """
        Retrieve a list of installed licenses

        :returns: A list of licenses
        :rtype: list[cterasdk.common.object.Object]
        """
        return self._core.api.get('/portalLicenses')

    def add(self, *keys):
        """
        Add license keys

        :param list[str] keys: List of license keys
        """
        logging.getLogger().info('Adding license(s)')
        param = Object()
        param.keys = list(keys)
        self._core.api.execute('', 'addLicenses', param)
        logging.getLogger().info('License(s) added')

    def remove(self, *keys):
        """
        Remove license keys

        :returns: A list of installed licenses
        :rtype: list[cterasdk.common.object.Object]
        """
        licenses = self.all()
        param = [license for license in licenses if license.key not in keys and license.originalKey not in keys]
        if len(param) != len(licenses):
            logging.getLogger().info('Updating licenses.')
            response = self._core.api.put('/portalLicenses', param)
            logging.getLogger().info('Licenses updated.')
            return response
        return None



import logging

from .enum import Mode
from ..exceptions import CTERAException
from .base_command import BaseCommand


class FTP(BaseCommand):
    """ Edge Filer FTP configuration APIs """

    def get_configuration(self):
        """
        Get the current FTP configuration

        :return cterasdk.common.object.Object:
        """
        return self._edge.api.get('/config/fileservices/ftp')

    def enable(self):
        """ Enable FTP """
        self._set_mode(True)

    def disable(self):
        """ Disable FTP """
        self._set_mode(False)

    def is_disabled(self):
        """ Check if the FTP server is disabled """
        return self._edge.api.get('/config/fileservices/ftp/mode') == Mode.Disabled

    def _set_mode(self, enabled):
        logging.getLogger().info('%s FTP server.', ('Enabling' if enabled else 'Disabling'))
        self._edge.api.put('/config/fileservices/ftp/mode', Mode.Enabled if enabled else Mode.Disabled)
        logging.getLogger().info('FTP server %s.', ('enabled' if enabled else 'disabled'))

    def modify(
            self,
            allow_anonymous_ftp=None,
            anonymous_download_limit=None,
            anonymous_ftp_folder=None,
            banner_message=None,
            max_connections_per_ip=None,
            require_ssl=None):
        """
        Modify the FTP Configuration. Parameters that are not passed will not be affected

        :param bool,optional allow_anonymous_ftp: Enable/Disable anonymous FTP downloads
        :param int,optional anonymous_download_limit:
         Limit download bandwidth of anonymous connection in KB/sec per connection. 0 for unlimited
        :param str,optional anonymous_ftp_folder: Anonymous FTP Directory
        :param str,optional banner_message: FTP Banner Message
        :param int,optional max_connections_per_ip: Maximum Connections per Client
        :param bool,optional require_ssl: If Ture, allow only SSL/TLS connections
        """
        config = self.get_configuration()
        if config.mode != Mode.Enabled:
            raise CTERAException("FTP must be enabled in order to modify its configuration")
        if anonymous_download_limit is not None:
            config.AnonymousDownloadLimit = anonymous_download_limit
        if anonymous_ftp_folder is not None:
            config.AnonymousFTPFolder = anonymous_ftp_folder
        if allow_anonymous_ftp is not None:
            config.AllowAnonymousFTP = allow_anonymous_ftp
        if banner_message is not None:
            config.BannerMessage = banner_message
        if max_connections_per_ip is not None:
            config.MaxConnectionsPerIP = max_connections_per_ip
        if require_ssl is not None:
            config.RequireSSL = require_ssl
        self._edge.api.put('/config/fileservices/ftp', config)



import logging

from ..common import Object
from .base_command import BaseCommand


class Drive(BaseCommand):
    """
    Edge Filer Drive APIs
    """

    def get(self, name=None):
        """
        Get Drive. If a drive name was not passed as an argument, a list of all drives will be retrieved

        :param str,optional name: Name of the drive
        """
        return self._edge.api.get('/config/storage/disks' + ('' if name is None else ('/' + name)))

    def get_status(self, name=None):
        """
        Get drive status. If a drive name was not passed as an argument, a list of all drives will be retrieved

        :param str name: Name of the drive
        """
        return self._edge.api.get('/status/storage/disks' + ('' if name is None else ('/' + name)))

    def format(self, name):
        """
        Format a drive

        :param str name: The name of the drive to format
        """
        param = Object()
        param.name = name

        self._edge.api.execute("/proc/storage", "format", param)

        logging.getLogger().info('Formatting drive. %s', {'drive': name})

    def format_all(self):
        """ Format all drives """
        drives = self._edge.api.get('/status/storage/disks')

        for drive in drives:
            self.format(drive.name)



import logging

from ..common import Object
from .enum import Mode, CIFSPacketSigning, SMBProtocol
from ..exceptions import CTERAException, InputError
from .base_command import BaseCommand


class SMB(BaseCommand):
    """ Edge Filer SMB configuration APIs """

    def enable(self):
        """ Enable SMB """
        logging.getLogger().info('Enabling SMB server.')
        self._edge.api.put('/config/fileservices/cifs/mode', Mode.Enabled)
        logging.getLogger().info('SMB server enabled.')

    def enable_abe(self):
        """ Enable ABE """
        logging.getLogger().info('Enabling ABE.')
        self._edge.api.put('/config/fileservices/cifs/hideUnreadable', True)
        logging.getLogger().info('Access Based Enumeration (ABE) enabled.')

    def disable_abe(self):
        """ Disable ABE """
        logging.getLogger().info('Disabling ABE.')
        self._edge.api.put('/config/fileservices/cifs/hideUnreadable', False)
        logging.getLogger().info('Access Based Enumeration (ABE) disabled.')

    def set_packet_signing(self, packet_signing):
        """
        Set Packet signing

        :param cterasdk.edge.enum.CIFSPacketSigning packet_signing: Packet signing type
        """
        self._verify_packet_signing_parameter(packet_signing)
        logging.getLogger().info('Updating SMB packet signing configuration.')
        try:
            self._edge.api.put('/config/fileservices/cifs/packetSigning', packet_signing)
            logging.getLogger().info('SMB packet signing configuration updated. %s', {'packet_signing': packet_signing})
        except CTERAException as error:
            logging.getLogger().error('Failed to update SMB packet signing configuration.')
            raise CTERAException('Invalid packet signing co', error)

    def disable(self):
        """ Disable SMB """
        logging.getLogger().info('Disabling SMB server.')
        self._edge.api.put('/config/fileservices/cifs/mode', Mode.Disabled)
        logging.getLogger().info('SMB server disabled.')

    def restart(self):
        self.disable()
        self.enable()

    def get_configuration(self):
        """
        Get current SMB Configuration

        :return cterasdk.common.object.Object: SMB configuration
        """
        cifs = self._edge.api.get('/config/fileservices/cifs')
        obj = Object()
        obj.mode = cifs.mode
        if obj.mode == Mode.Enabled:
            obj.packet_signing = cifs.packetSigning
            obj.idle_disconnect_time = cifs.idleDisconnectTime
            obj.compatibility_mode = cifs.compatibilityMode
            obj.cifs_unix_extensions = cifs.cifsUnixExtensions
            obj.abe_enabled = cifs.hideUnreadable
            obj.min_client_protocol = cifs.minClientProtocol
            obj.max_client_protocol = cifs.maxClientProtocol
            obj.min_server_protocol = cifs.minServerProtocol
            obj.max_server_protocol = cifs.maxServerProtocol
        return obj

    def modify(
            self,
            packet_signing=None,
            idle_disconnect_time=None,
            compatibility_mode=None,
            unix_extensions=None,
            abe_enabled=None,
            min_client_protocol=None,
            max_client_protocol=None,
            min_server_protocol=None,
            max_server_protocol=None):
        """
        Modify the current SMB Configuration. Parameters that are not passed will not be affected

        :param cterasdk.edge.enum.CIFSPacketSigning packet_signing,optional: Packet signing type
        :param int,optional idle_disconnect_time: Client idle disconnect timeout
        :param bool,optional compatibility_mode: Enable/Disable compatibility mode
        :param bool,optional unix_extensions: Enable/Disable unix extensions
        :param bool,optional abe_enabled: Enable/Disable ABE
        :param cterasdk.edge.enum.SMBProtocol,optional min_client_protocol: Minimum client protocol version
        :param cterasdk.edge.enum.SMBProtocol,optional max_client_protocol: Maximum client protocol version
        :param cterasdk.edge.enum.SMBProtocol,optional min_server_protocol: Minimum server protocol version
        :param cterasdk.edge.enum.SMBProtocol,optional max_server_protocol: Maximum server protocol version
        """
        cifs = self._edge.api.get('/config/fileservices/cifs')
        if cifs.mode != Mode.Enabled:
            raise CTERAException("SMB must be enabled in order to modify its configuration")
        if packet_signing is not None:
            self._verify_packet_signing_parameter(packet_signing)
            cifs.packetSigning = packet_signing
        if idle_disconnect_time is not None:
            cifs.idleDisconnectTime = idle_disconnect_time
        if compatibility_mode is not None:
            cifs.compatibilityMode = compatibility_mode
        if unix_extensions is not None:
            cifs.cifsUnixExtensions = unix_extensions
        if abe_enabled is not None:
            cifs.hideUnreadable = abe_enabled
        if min_client_protocol is not None:
            self._verify_smb_protocol_version_parameter(min_client_protocol)
            cifs.minClientProtocol = min_client_protocol
        if max_client_protocol is not None:
            self._verify_smb_protocol_version_parameter(max_client_protocol)
            cifs.maxClientProtocol = max_client_protocol
        if min_server_protocol is not None:
            self._verify_smb_protocol_version_parameter(min_server_protocol)
            cifs.minServerProtocol = min_server_protocol
        if max_server_protocol is not None:
            self._verify_smb_protocol_version_parameter(max_server_protocol)
            cifs.maxServerProtocol = max_server_protocol
        try:
            self._edge.api.put('/config/fileservices/cifs', cifs)
            logging.getLogger().info('SMB configuration updated.')
        except CTERAException as error:
            msg = 'Failed to update SMB configuration.'
            logging.getLogger().error(msg)
            raise CTERAException(message=msg, instace=error)

    @staticmethod
    def _verify_smb_protocol_version_parameter(smb_protocol_version):
        options = [v for k, v in SMBProtocol.__dict__.items() if not k.startswith('_')]
        if smb_protocol_version not in options:
            raise InputError('Invalid SMB protocol version', smb_protocol_version, options)

    @staticmethod
    def _verify_packet_signing_parameter(packet_signing):
        options = [v for k, v in CIFSPacketSigning.__dict__.items() if not k.startswith('_')]
        if packet_signing not in options:
            raise InputError('Invalid packet signing option', packet_signing, options)



import logging
from ...exceptions import CTERAException
from . import common


def mkdir(edge, path):
    directory = path.fullpath()
    logging.getLogger().info('Creating directory. %s', {'path': directory})
    try:
        edge.io.mkdir(path.fullpath())
    except CTERAException as error:
        try:
            common.raise_for_status(error.response.message.msg, directory)
        except common.ItemExists:
            logging.getLogger().info('Directory already exists. %s', {'path': directory})
    logging.getLogger().info('Directory created. %s', {'path': directory})
    return directory


def makedirs(edge, path):
    directories = path.parts()
    for i in range(1, len(directories) + 1):
        path = common.get_object_path(path.base, '/'.join(directories[:i]))
        try:
            mkdir(edge, path)
        except common.Forbidden:
            pass


def copy(edge, path, destination=None, overwrite=False):
    destination = destination.joinpath(path.name()).fullpath()
    logging.getLogger().info('Copying. %s', {'from': path.fullpath(), 'to': destination})
    edge.io.copy(path.fullpath(), destination, overwrite=overwrite)
    logging.getLogger().info('Copied. %s', {'from': path.fullpath(), 'to': destination})


def move(edge, path, destination=None, overwrite=False):
    destination = destination.joinpath(path.name()).fullpath()
    logging.getLogger().info('Moving. %s', {'from': path.fullpath(), 'to': destination})
    edge.io.move(path.fullpath(), destination, overwrite=overwrite)
    logging.getLogger().info('Moved. %s', {'from': path.fullpath(), 'to': destination})


def remove(edge, *paths):
    for path in paths:
        logging.getLogger().info('Deleting object. %s', {'path': path.fullpath()})
        edge.io.delete(path.fullpath())
        logging.getLogger().info('Object deleted. %s', {'path': path.fullpath()})



from . import async_requests
from ..convert import tojsonstr, toxmlstr, fromjsonstr, fromxmlstr


class Serializers:
    JSON = tojsonstr
    XML = toxmlstr
    FormData = async_requests.FormData


class Deserializers:
    JSON = fromjsonstr
    XML = fromxmlstr


class MultipartForm:
    """Multipart Request Form"""

    def __init__(self):
        self._data = async_requests.FormData()

    def add(self, name, value, filename=None, content_transfer_encoding=None):
        self._data.add_field(name=name, value=value, filename=filename,
                             content_type='multipart/form-data', content_transfer_encoding=content_transfer_encoding)

    @property
    def data(self):
        return self._data



import logging

from .types import UserAccount
from .base_command import BaseCommand


class Credentials(BaseCommand):
    """
    Portal Credential Management APIs

    :ivar cterasdk.core.credentials.S3 s3: Object holding the Portal S3 Credential Management APIs.
    """

    def __init__(self, portal):
        super().__init__(portal)
        self.s3 = S3(self._core)


class S3(BaseCommand):
    """S3 Credential Management APIs"""

    def _user_account(self, user_account=None):
        return user_account if user_account else UserAccount(self._core.session().user.name)

    def all(self, user_account=None):
        """
        List Credentials

        :param cterasdk.core.types.UserAccount,optional user_account: User account
        """
        user_account = self._user_account(user_account)
        param = self._core.users.get(user_account, ['uid']).uid
        logging.getLogger().info('Listing Credentials. %s', {'user': str(user_account)})
        return self._core.api.execute('', 'getApiKeys', param)

    def create(self, user_account=None):
        """
        Create Credential

        :param cterasdk.core.types.UserAccount,optional user_account: User account
        """
        user_account = self._user_account(user_account)
        param = self._core.users.get(user_account, ['uid']).uid
        logging.getLogger().info('Creating Credential. %s', {'type': 's3', 'user': str(user_account)})
        response = self._core.api.execute('', 'createApiKey', param)
        logging.getLogger().info('Credetial created. %s', {'type': 's3', 'user': str(user_account)})
        return response

    def delete(self, access_key_id, user_account=None):
        """
        Delete Credential

        :param str access_key_id: Access Key ID
        :param cterasdk.core.types.UserAccount,optional user_account: User account
        """
        user_account = self._user_account(user_account)
        for key in self.all(user_account):
            if key.accessKey == access_key_id:
                logging.getLogger().info('Deleting Credential. %s', {'type': 's3'})
                response = self._core.api.execute('', 'deleteApiKey', key.uid)
                logging.getLogger().info('Credetial deleted. %s', {'type': 's3'})
                return response
        logging.getLogger().warning('Could not find access key. %s', {'key': access_key_id})
        return None



import json
import logging
import aiohttp


def default():

    async def on_request_start(session, context, params):
        # pylint: disable=unused-argument
        param = {
            'request': {
                'method': params.method,
                'url': str(params.url),
                'headers': [dict(params.headers)] if params.headers else []
            }
        }
        logging.getLogger().debug('Starting request. %s', serialize(param))

    async def on_request_redirect(session, context, params):
        # pylint: disable=unused-argument
        param = {
            'redirect': {
                'source': str(params.response.real_url),
                'destination': params.response.headers['Location']
            }
        }
        logging.getLogger().debug('Starting redirect. %s', serialize(param))

    async def on_request_end(session, context, params):
        # pylint: disable=unused-argument
        param = {
            'request': {
                'method': params.response.method,
                'url': str(params.response.real_url),
                'headers': [dict(params.response.request_info.headers)]
            },
            'response': {
                'status': params.response.status,
                'reason': params.response.reason,
                'headers': [dict(params.response.headers)]
            }
        }
        logging.getLogger().debug('Ended request. %s', serialize(param))

    trace_config = aiohttp.TraceConfig()
    trace_config.on_request_start.append(on_request_start)
    trace_config.on_request_redirect.append(on_request_redirect)
    trace_config.on_request_end.append(on_request_end)

    return trace_config


def serialize(param):
    return json.dumps(param, indent=5)



import logging

from .base_command import BaseCommand
from ..common import Object


class Messaging(BaseCommand):
    """
    Portal Messaging Service Management APIs
    """

    def is_active(self):
        """
        Check if messaging service is Active
        """
        return self._core.api.get('/microservices/messaging/globalStatus').status == 'Active'

    def get_status(self):
        """
        Retrieve the global status of messaging service
        """
        res = self._core.api.get('/microservices/messaging/globalStatus')
        return res

    def get_servers_status(self):
        """
        Retrieve the status of the messaging servers
        """
        return {f'{srv.server.name}: "{srv.serverStatus.status}"' for srv in self._core.api.get('/microservices/messaging').currentNodes}

    def add(self, servers):
        """
        Add messaging servers to cluster

        :param list[str] servers: Server names (number of allowed servers: 1 or 3)

        """
        nodes = []
        messaging_obj = self._core.api.get('/microservices/messaging')
        if messaging_obj.globalStatus.canAddServers:
            for node in messaging_obj.availableNodes:
                if node.server.name in servers and node.canAssignAsMessaging.allowed:
                    param = Object()
                    param._class = 'CurrentMessagingNode'  # pylint: disable=protected-access
                    param.server = node.server
                    nodes.append(param)
            if len(nodes) in messaging_obj.globalStatus.validServerNumber:
                response = self._core.api.put('microservices/messaging/currentNodes', nodes)
                logging.getLogger().info('Nodes added to cluster successfully')
                return response
            logging.getLogger().error('Wrong number of servers. expected:"1" or "3", %s', {'given': len(servers)})
        else:
            logging.getLogger().error('Can not add new servers: %s', {messaging_obj.globalStatus.cantAddServersReason})
        return None



import logging
import cterasdk.settings

from ..lib.task_manager_base import TaskError
from .licenses import Licenses
from ..lib import ask, track
from ..common import Object
from ..exceptions import CTERAException, InputError
from . import enum
from .base_command import BaseCommand
from .types import TCPService


class Services(BaseCommand):
    """ Edge Filer Cloud Services configuration APIs """
    _UNTRUSTED_CERTIFICATE_ERRORS = [
        'UntrustedCert',
        'UntrustedCA',
        'UntrustedCertExpired',
        'UntrustedCertSelfSigned',
        'UntrustedCertDNS'
    ]

    def __init__(self, edge):
        super().__init__(edge)
        self._trust_cert = {}

    def get_status(self):
        """
        Retrieve the cloud services connection status
        """
        status = self._edge.api.get('/status/services')
        connection = Object()
        connection.connected = status.CTERAPortal.connectionState == enum.ServicesConnectionState.Connected
        if connection.connected:
            connection.ipaddr = status.CTERAPortal.connectedAddress
            connection.user = status.userDisplayName
            connection.server_version = status.portalVersion
            connection.server_address = status.CTERAPortal.serverList[0].name
            connection.last_connected_at = status.CTERAPortal.establishedTime
        return connection

    def connected(self):
        """
        Check if the Edge Filer is connected to CTERA Portal
        """
        return self._edge.api.get('/status/services/CTERAPortal/connectionState') == enum.ServicesConnectionState.Connected

    def _before_connect_to_services(self, ctera_license, server):
        Services._validate_license(ctera_license)
        if self._edge.network.proxy.is_enabled():
            logging.getLogger().debug('Skipping TCP connection verification over port 995.')
        else:
            self._check_cttp_traffic(address=server)

    def connect(self, server, user, password, ctera_license=enum.License.EV16):
        """
        Connect to a Portal.\n
        The connect method will first validate the `license` argument,
         ensure the Edge Filer can establish a TCP connection over port 995 to the Portal and
         verify the Portal does not require device activation via code.
         TCP connection verification over port 995 is skipped when the Edge Filer is configured to use a proxy.

        :param str server: Address of the Portal
        :param str user: User for the Portal connection
        :param str password: Password for the Portal connection
        :param cterasdk.edge.enum.License,optional ctera_license: CTERA License, defaults to cterasdk.edge.enum.License.EV16
        """
        self._before_connect_to_services(ctera_license, server)
        self._check_connection(server)

        param = Object()
        param.server = server
        param.user = user
        param.password = password
        param.trustCertificate = self._trust_cert.get(server, False)
        self._connect_to_services(param, ctera_license)

    def activate(self, server, user, code, ctera_license=enum.License.EV16):
        """
        Enroll the Edge Filer with CTERA Portal using an activation code

        :param str server: Address of the Portal
        :param str user: User for  the Portal connection
        :param str code: Activation code for the Portal connection
        :param cterasdk.edge.enum.License,optional ctera_license: CTERA License, defaults to cterasdk.edge.enum.License.EV16
        """
        self._before_connect_to_services(ctera_license, server)

        param = Object()
        param.server = server
        param.user = user
        param.activationCode = code
        param.trustCertificate = self._trust_cert.get(server, False)
        self._connect_to_services(param, ctera_license)

    def reconnect(self):
        """ Reconnect to the Portal """
        self._edge.api.execute("/status/services", "reconnect", None)

    def disconnect(self):
        """ Disconnect from the Portal """
        self._edge.api.put('/config/services', None)

    def sso_enabled(self):
        """
        Is SSO connection enabled

        :return bool: True if SSO connection is enabled, else False
        """
        return self._edge.api.get('/config/gui/adminRemoteAccessSSO')

    def enable_sso(self):
        """ Enable SSO connection """
        self._set_sso(True)

    def disable_sso(self):
        """ Disable SSO connection """
        self._set_sso(False)

    def _set_sso(self, sso_state):
        logging.getLogger().info('%s single sign-on from CTERA Portal.', ('Enabling' if sso_state else 'Disabling'))
        self._edge.api.put('/config/gui/adminRemoteAccessSSO', sso_state)
        logging.getLogger().info('Single sign-on %s.', ('enabled' if sso_state else 'disabled'))

    def _connect_to_services(self, param, ctera_license):
        task = self._attach(param)
        try:
            self._edge.tasks.wait(task)
            track(self._edge, '/status/services/CTERAPortal/connectionState', [enum.ServicesConnectionState.Connected],
                  [enum.ServicesConnectionState.ResolvingServers, enum.ServicesConnectionState.Connecting,
                   enum.ServicesConnectionState.Attaching, enum.ServicesConnectionState.Authenticating],
                  [], [enum.ServicesConnectionState.Disconnected], 20, 1)
            logging.getLogger().info("Connected to Portal.")
        except TaskError as error:
            description = error.task.description
            logging.getLogger().error("Connection failed. Reason: %s", description)
            raise CTERAException("Connection failed", None, reason=description)
        self._edge.licenses.apply(ctera_license)

    @staticmethod
    def _validate_license(ctera_license):
        try:
            Licenses.infer(ctera_license)
        except InputError as error:
            logging.getLogger().error('Connection failed. Invalid license type. %s', {'license': ctera_license})
            raise error

    def _check_cttp_traffic(self, address, port=995):
        tcp_connect_result = self._edge.network.tcp_connect(TCPService(address, port))
        if not tcp_connect_result.is_open:
            logging.getLogger().error("Unable to establish connection over port %s", str(tcp_connect_result.port))
            raise ConnectionError(f'Unable to establish CTTP connection {tcp_connect_result.host}:{tcp_connect_result.port}')

    def _check_connection(self, server):
        param = Object()
        param.server = server
        param.trustCertificate = self._trust_cert.get(server, False)
        obj = self._edge.api.execute('/status/services', 'isWebSsoEnabled', param)
        if not Services._check_web_sso(obj):
            self._handle_untrusted_cert(server, obj)

    @staticmethod
    def _check_web_sso(obj):
        try:
            if obj.result.hasWebSSO:
                message = "Connection failed. You must activate this Edge Filer using an activation code."
                logging.getLogger().error(message)
                raise CTERAException(message)

            if not obj.result.hasWebSSO and obj.rc == "connectOK":
                return True
            raise CTERAException("Connection failed", None, reason=obj.rc)
        except AttributeError:
            pass

        return False

    def _handle_untrusted_cert(self, server, obj):
        try:
            if obj.rc in Services._UNTRUSTED_CERTIFICATE_ERRORS:
                proceed = False
                if cterasdk.settings.sessions.management.edge.services.ssl == 'prompt':
                    logging.getLogger().warning(msg=obj.msg)
                    proceed = ask(f"Connect {self._edge.host()} to {server}?")
                if cterasdk.settings.sessions.management.edge.services.ssl is False or proceed:
                    self._trust_cert[server] = True
                    return self._check_connection(server)
        except AttributeError:
            pass
        return False

    def _attach(self, param):
        logging.getLogger().info("Connecting to Portal. %s", {'server': param.server, 'user': param.user})
        obj = self._edge.api.execute("/status/services", "attachAndSave", param)
        return obj.id



__all__ = [
    'afp',
    'aio',
    'array',
    'audit',
    'backup',
    'cache',
    'cli',
    'config',
    'connection',
    'directoryservice',
    'drive',
    'enum',
    'files',
    'ftp',
    'groups',
    'licenses',
    'login',
    'logs',
    'mail',
    'network',
    'nfs',
    'ntp',
    'power',
    'query',
    'remote',
    'rsync',
    'ransom_protect',
    'services',
    'session',
    'shares',
    'shell',
    'smb',
    'support',
    'sync',
    'syslog',
    'telnet',
    'timezone',
    'users',
    'volumes',
]



import logging

from .base_command import BaseCommand
from .types import GroupAccount
from ..exceptions import CTERAException, ObjectNotFoundException
from ..common import Object
from ..common import union
from . import query


class Groups(BaseCommand):
    """
    Portal Groups Management APIs
    """

    default = ['name']

    @staticmethod
    def _build_resource_url(group_account):
        return f'/localGroups/{group_account.name}' if group_account.is_local \
            else f'/domains/{group_account.directory}/adGroups/{group_account.name}'

    def _get_entire_object(self, group_account):
        ref = Groups._build_resource_url(group_account)
        try:
            return self._core.api.get(ref)
        except CTERAException as error:
            raise CTERAException('Failed to retrieve group', error)

    def get(self, group_account, include=None):
        """
        Get a group account

        :param cterasdk.core.types.GroupAccount group_account: Group account, including the group directory and name
        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: The group account, including the requested fields
        """
        baseurl = Groups._build_resource_url(group_account)
        include = union(include or [], Groups.default)
        include = ['/' + attr for attr in include]
        group_object = self._core.api.get_multi(baseurl, include)
        if group_object.name is None:
            raise ObjectNotFoundException('Could not find group', baseurl, group_directory=group_account.directory, name=group_account.name)
        return group_object

    def list_local_groups(self, include=None):
        """
        List all local groups

        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: Iterator for all local groups
        :rtype: cterasdk.lib.iterator.Iterator
        """
        include = union(include or [], Groups.default)
        param = query.QueryParamBuilder().include(include).build()
        return query.iterator(self._core, '/localGroups', param)

    def list_domain_groups(self, domain, include=None):
        """
        List all the groups in the domain

        :param str domain: Domain name
        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: Iterator for all the domain groups
        :rtype: cterasdk.lib.iterator.Iterator
        """
        include = union(include or [], Groups.default)
        param = query.QueryParamBuilder().include(include).build()
        return query.iterator(self._core, f'/domains/{domain}/adGroups', param)

    def _members_reference(self, users):
        return [self._core.users.get(user, include=['baseObjectRef']).baseObjectRef for user in users]

    def add(self, name, description=None, members=None):
        """
        Create a local group

        :param str name: Group name
        :param str,optional description: Group description
        :param list(cterasdk.core.types.UserAccount),optional members: List of user group members
        """
        param = Object()
        param._classname = 'AddGroupParam'  # pylint: disable=protected-access
        param.groupData = Object()
        param.groupData._classname = 'PortalGroup'  # pylint: disable=protected-access
        param.groupData.name = name
        param.groupData.description = description
        if members:
            param.members = self._members_reference(members)

        logging.getLogger().info('Creating group. %s', {'group': name})
        response = self._core.api.execute('', 'addGroup', param)
        logging.getLogger().info('Group created. %s', {'group': name})

        return response

    def modify(self, current_groupname, new_groupname=None, description=None):
        """
        Modify a local group

        :param str current_groupname: The current group name
        :param str,optional new_groupname: New group name
        :param str,optional description: Group description
        """
        group_account = GroupAccount(current_groupname)
        group = self._get_entire_object(group_account)

        if new_groupname:
            group.name = new_groupname
        if description:
            group.description = description

        param = Object()
        param._classname = 'UpdateGroupParam'  # pylint: disable=protected-access
        param.groupData = group
        param.membersToAdd = []
        param.membersToDelete = []

        try:
            response = self._core.api.execute(f'/localGroups/{current_groupname}', 'updateGroup', param)
            logging.getLogger().info("Group modified. %s", {'group_name': group.name})
            return response
        except CTERAException as error:
            logging.getLogger().error("Failed to modify group.")
            raise CTERAException('Failed to modify group', error)

    def get_members(self, group_account):
        """
        Get group members

        :param cterasdk.core.types.GroupAccount group_account: Group account
        """
        param = query.QueryParamBuilder().startFrom(0).build()
        return list(query.iterator(self._core, Groups._build_resource_url(group_account), param, 'getMembers'))

    def add_members(self, group_account, members):
        """
        Add group members

        :param cterasdk.core.types.GroupAccount group_account: Group account
        :param list(cterasdk.core.types.UserAccount) members: Users
        """
        current_members = self.get_members(group_account)
        new_members = self._members_reference(members)
        param = Object()
        param.groupData = self._get_entire_object(group_account)
        param.membersToAdd = list(set(new_members) - set(current_members))
        param.membersToDelete = []
        return self._core.api.execute(f'/localGroups/{group_account.name}', 'updateGroup', param)

    def remove_members(self, group_account, members):
        """
        Remove group members

        :param cterasdk.core.types.GroupAccount group_account: Group account
        :param list(cterasdk.core.types.UserAccount) members: Users
        """
        current_members = self.get_members(group_account)
        to_remove = self._members_reference(members)
        param = Object()
        param.groupData = self._get_entire_object(group_account)
        param.membersToAdd = []
        param.membersToDelete = list(set(current_members).intersection(set(to_remove)))
        return self._core.api.execute(f'/localGroups/{group_account.name}', 'updateGroup', param)

    def delete(self, group_account):
        """
        Delete a group

        :param cterasdk.core.types.GroupAccount group_account: Group account
        """
        logging.getLogger().info('Deleting group. %s', {'group': str(group_account.name)})
        baseurl = f'/localGroups/{group_account.name}' if group_account.is_local \
            else f'/domains/{group_account.directory}/adGroups/{group_account.name}'
        response = self._core.api.execute(baseurl, 'delete', True)
        logging.getLogger().info('Group deleted. %s', {'group': str(group_account.name)})

        return response



import copy

from ..common import Object
from ..common.utils import Version


class SessionStatus:
    Initializing = 'Initializing'
    Inactive = 'Inactive'
    Active = 'Active'


class SessionUser(Object):
    def __init__(self, name):
        self.name = name


class SessionHostType:
    Edge = 'Edge Filer'
    Portal = 'Portal'
    Agent = 'Agent'


class SessionBase(Object):

    def __init__(self, host, host_type):
        self.host = host
        self.host_type = host_type
        self.status = SessionStatus.Inactive
        self.user = None
        self.version = None

    def set_version(self, version):
        self.version = Version(version)

    def start_local_session(self, CTERA_Host):
        self.status = SessionStatus.Initializing
        self._do_start_local_session(CTERA_Host)
        self.status = SessionStatus.Active

    def _do_start_local_session(self, CTERA_Host):
        raise NotImplementedError("Subclass must implement the _do_start_local_session method")

    def terminate(self):
        self._do_terminate()
        self.status = SessionStatus.Inactive
        self.user = None

    def _do_terminate(self):
        raise NotImplementedError("Subclass must implement the _do_terminate method")

    def initializing(self):
        return self.status == SessionStatus.Initializing

    def authenticated(self):
        return self.status == SessionStatus.Active

    @property
    def active(self):
        return self.status == SessionStatus.Active

    def tenant(self):
        return self.user.tenant if self.user else None

    def whoami(self):
        session = copy.deepcopy(self)
        session.version = str(session.version)
        print(session)



import logging
from .base_command import BaseCommand


class Timezone(BaseCommand):
    """ Edge Filer Timezone configuration """

    def get_timezone(self):
        """
        Get the timezone of the Edge Filer

        :return str: The timezone of the Edge Filer
        """
        return self._edge.api.get('/config/time/TimeZone')

    def set_timezone(self, timezone):
        """
        Set Timezone

        :param str timezone: New timezone to set
        """
        logging.getLogger().info("Updating timezone. %s", {'timezone': timezone})

        self._edge.api.put('/config/time/TimeZone', timezone)

        logging.getLogger().info("Timezone updated. %s", {'timezone': timezone})



from . import async_requests
from ..common import utils


class CookieJar:

    def __init__(self, cookies):
        self._cookies = cookies

    def get(self, key):
        return self._cookies.get(key)

    def update(self, cookies, response_url):
        self._cookies.update_cookies(cookies, response_url)


class PersistentHeaders:
    """Headers to include in every request"""

    def __init__(self):
        self._headers = {}

    @property
    def all(self):
        return self._headers

    def update_headers(self, headers):
        self._headers.update(headers)


class BaseClient:
    """Base Client"""

    def __init__(self, builder=None, async_session=None, authenticator=None, **session_settings):
        """
        Initialize a Client

        :param builder: Endpoint builder.
        :param ,optional async_session: Re-use an asynchronous session.
        :param ,optional authenticator: Authenticator function.
        """
        self._headers = PersistentHeaders()
        self._authenticator = authenticator
        self._builder = builder
        self._async_session = async_session if async_session else async_requests.Session(**session_settings)

    @property
    def cookies(self):
        return CookieJar(self._async_session.cookies)

    @property
    def headers(self):
        return self._headers

    def join_headers(self, request):
        request.kwargs['headers'] = utils.merge(request.kwargs.get('headers', None), self.headers.all)
        return request

    @property
    def baseurl(self):
        return self._builder()

    def __str__(self):
        return f"({self.__class__.__name__} client at {hex(hash(self))}, baseurl={self.baseurl})"

    async def shutdown(self):
        await self._async_session.shutdown()


class BaseResponse:
    """Base Response Object"""

    def __init__(self, response):
        self._response = response

    @property
    def method(self):
        return self._response.method

    @property
    def ok(self):
        return self._response.ok

    @property
    def reason(self):
        return self._response.reason

    @property
    def url(self):
        return self._response.url

    @property
    def real_url(self):
        return self._response.real_url

    @property
    def status(self):
        return self._response.status

    @property
    def headers(self):
        return self._response.headers

    def raise_for_status(self):
        return self._response.raise_for_status()



import logging

from .base_command import BaseCommand
from ..common import Object, utf8_decode
from ..lib import X509Certificate, PrivateKey
from . import query


class KMS(BaseCommand):
    """
    External Key Management APIs
    :ivar cterasdk.core.kms.KMS servers: Object holding the Portal External Key Management Server APIs
    """

    def __init__(self, portal):
        super().__init__(portal)
        self.servers = KMSServers(self._core)

    def settings(self):
        """
        Get Key Management Service Settings
        """
        return self._core.api.get('/settings/keyManagerSettings')

    def status(self):
        """
        Get Key Management Service Status
        """
        return self._core.api.execute('', 'getKeyManagerGlobalStatus')

    def enable(self, private_key, client_certificate, server_certificate, expiration=None, timeout=None, port=None):
        """
        Enable Key Management Service

        :param str private_key: The PEM-encoded private key, or a path to the PEM-encoded private key file
        :param str client_certificate: The PEM-encoded client certificate, or a path to the certificate file
        :param str server_certificate: The PEM-encoded KMS server certificate, or a path to the certificate file
        :param int,optional expiration: Key expiration in days, defaults to ``365``.
        :param int,optional timeout: Connection timeout in seconds, defaults to ``2``
        :param int,optional port: Key server port, defaults to ``5696``
        """
        param = self._core.api.defaults('KeyManagerSettings')
        param.expiration = expiration if expiration else 365
        param.integration.connectionSettings.timeout = timeout if timeout else 2
        param.integration.connectionSettings.port = port if port else 5696
        param.integration.tlsDetails = self._TLS_details(private_key, client_certificate, server_certificate)
        logging.getLogger().info('Enabling Key Management Service')
        response = self._core.api.put('/settings/keyManagerSettings', param)
        logging.getLogger().info('Key Management Service enabled')
        return response

    @staticmethod
    def _TLS_details(private_key, client_certificate, server_certificate):
        param = Object()
        param._classname = 'TLSDetails'  # pylint: disable=protected-access
        param.files = Object()
        param.files._classname = 'TLSFiles'  # pylint: disable=protected-access
        param.files.clientCert = utf8_decode(X509Certificate.load_certificate(client_certificate).pem_data)
        param.files.privateKey = utf8_decode(PrivateKey.load_private_key(private_key).pem_data)
        param.files.rootCACert = utf8_decode(X509Certificate.load_certificate(server_certificate).pem_data)
        return param

    def disable(self):
        """
        Disable Key Management Service
        """
        logging.getLogger().info('Disabling Key Management Service')
        response = self._core.api.execute('', 'removeKeyManagementService')
        logging.getLogger().info('Key Management Service disabled')
        return response

    def modify(self, private_key=None, client_certificate=None, server_certificate=None, expiration=None, timeout=None, port=None):
        """
        Modify Key Management Service Settings

        :param str,optional private_key: The PEM-encoded private key, or a path to the PEM-encoded private key file
        :param str,optional client_certificate: The PEM-encoded client certificate, or a path to the certificate file
        :param str,optional server_certificate: The PEM-encoded KMS server certificate, or a path to the certificate file
        :param int,optional expiration: Key expiration in days, defaults to ``365``.
        :param int,optional timeout: Connection timeout in seconds, defaults to ``2``
        :param int,optional port: Key server port, defaults to ``5696``
        """

        settings = self.settings()
        if client_certificate:
            settings.integration.tlsDetails.files.clientCert = utf8_decode(X509Certificate.load_certificate(client_certificate).pem_data)
        if private_key:
            settings.integration.tlsDetails.files.privateKey = utf8_decode(PrivateKey.load_private_key(private_key).pem_data)
        if server_certificate:
            settings.integration.tlsDetails.files.rootCACert = utf8_decode(X509Certificate.load_certificate(server_certificate).pem_data)
        if expiration:
            settings.expiration = expiration
        if timeout:
            settings.integration.connectionSettings.timeout = timeout
        if port:
            settings.integration.connectionSettings.port = port

        logging.getLogger().info('Updating Key Management Service settings')
        response = self._core.api.put('/settings/keyManagerSettings', settings)
        logging.getLogger().info('Updated Key Management Service settings')
        return response


class KMSServers(BaseCommand):
    """ External Key Management Server APIs """

    def get(self, name):
        """
        Retrieve a key-server

        :param str name: Key-server name
        """
        return self._core.api.get(f'/keyManagerServers/{name}')

    def all(self):
        """
        List Key Management Servers
        """
        param = query.QueryParamBuilder().startFrom(0).countLimit(25).orFilter(True).build()
        return query.iterator(self._core, '/keyManagerServers', param)

    def add(self, name, ipaddr):
        """
        Add a key-server

        :param str name: Key-server name
        :param str ipaddr: Key-server IP address
        """
        param = Object()
        param._classname = 'KeyManagerServer'  # pylint: disable=protected-access
        param.name = name
        param.host = ipaddr
        logging.getLogger().info('Adding Key Server. %s', {'name': name, 'host': ipaddr})
        response = self._core.api.add('/keyManagerServers', param)
        logging.getLogger().info('Key Server. %s Added', {'name': name, 'host': ipaddr})
        return response

    def modify(self, current_name, new_name):
        """
        Remove a key-server

        :param str current_name: Key-server current name
        :param str new_name: Key-server new name
        """
        key_server = self.get(current_name)
        key_server.name = new_name
        logging.getLogger().info("Modifying Key Server. %s", {'name': current_name})
        response = self._core.api.put(f'/keyManagerServers/{current_name}', key_server)
        logging.getLogger().info("Key Server modified. %s", {'name': current_name})
        return response

    def delete(self, name):
        """
        Remove a key-server

        :param str name: Key-server name
        """
        logging.getLogger().info('Deleting Key Server. %s', {'name': name})
        response = self._core.api.delete(f'/keyManagerServers/{name}')
        logging.getLogger().info('Key Server deleted. %s', {'name': name})
        return response



import logging

import cterasdk.settings
from ...exceptions import CTERAException
from ..base_command import BaseCommand
from . import io, common, shares, file_access


class FileBrowser(BaseCommand):

    def __init__(self, core):
        super().__init__(core)
        self._base = f'/{self._core.context}/webdav'
        self._file_access = file_access.FileAccess(self._core)

    def download(self, path, destination=None):
        """
        Download a file

        :param str path: Path
        :param str,optional destination:
         File destination, if it is a directory, the original filename will be kept, defaults to the default directory
        """
        return self._file_access.download(self.get_object_path(path), destination=destination)

    def download_as_zip(self, cloud_directory, files, destination=None):
        """
        Download a list of files and/or directories from a cloud folder as a ZIP file

        .. warning:: The list of files is not validated. The ZIP file will include only the existing  files and directories

        :param str cloud_directory: Path to the cloud directory
        :param list[str] files: List of files and/or directories in the cloud folder to download
        :param str,optional destination:
         File destination, if it is a directory, the original filename will be kept, defaults to the default directory
        """
        self._file_access.download_as_zip(self.get_object_path(cloud_directory), files, destination=destination)

    def listdir(self, path, depth=None, include_deleted=False):
        """
        List Directory

        :param str path: Path
        :param bool,optional include_deleted: Include deleted files, defaults to False
        """
        return io.listdir(self._core, self.get_object_path(path), depth=depth, include_deleted=include_deleted)

    def walk(self, path, include_deleted=False):
        """
        Walk Directory Contents

        :param str path: Path to walk
        :param bool,optional include_deleted: Include deleted files, defaults to False
        """
        return io.walk(self._core, self._base, path, include_deleted=include_deleted)

    def public_link(self, path, access='RO', expire_in=30):
        """
        Create a public link to a file or a folder

        :param str path: The path of the file to create a link to
        :param str,optional access: Access policy of the link, defaults to 'RO'
        :param int,optional expire_in: Number of days until the link expires, defaults to 30
        """
        return shares.create_public_link(self._core, self.get_object_path(path), access, expire_in)

    def copy(self, *paths, destination=None):
        """
        Copy one or more files or folders

        :param list[str] paths: List of paths
        :param str destination: Destination
        """
        if destination is None:
            raise ValueError('Copy destination was not specified.')
        return io.copy(self._core, *[self.get_object_path(path) for path in paths], destination=self.get_object_path(destination))

    def get_object_path(self, elements):
        return common.get_object_path(self.base, elements)

    @property
    def base(self):
        return self._base


class CloudDrive(FileBrowser):

    def upload(self, path, destination):
        """
        Upload a file

        :param str path: Local path
        :param str destination: Remote path
        """
        self._file_access.upload(path, self.get_object_path(destination))

    def mkdir(self, path):
        """
        Create a new directory

        :param str path: Directory path
        """
        return io.mkdir(self._core, self.get_object_path(path))

    def makedirs(self, path):
        """
        Create a directory recursively

        :param str path: Directory path
        """
        return io.makedirs(self._core, self.get_object_path(path))

    def rename(self, path, name):
        """
        Rename a file

        :param str path: Path of the file or directory to rename
        :param str name: The name to rename to
        """
        return io.rename(self._core, self.get_object_path(path), name)

    def delete(self, *paths):
        """
        Delete one or more files or folders

        :param str path: Path
        """
        return io.remove(self._core, *[self.get_object_path(path) for path in paths])

    def undelete(self, *paths):
        """
        Recover one or more files or folders

        :param str path: Path
        """
        return io.recover(self._core, *[self.get_object_path(path) for path in paths])

    def move(self, *paths, destination=None):
        """
        Move one or more files or folders

        :param list[str] paths: List of paths
        :param str destination: Destination
        """
        if destination is None:
            raise ValueError('Move destination was not specified.')
        return io.move(self._core, *[self.get_object_path(path) for path in paths], destination=self.get_object_path(destination))

    def get_share_info(self, path):
        """
        Get share settings and recipients

        :param str path: Path
        """
        return shares.get_share_info(self._core, self.get_object_path(path))

    def share(self, path, recipients, as_project=True, allow_reshare=True, allow_sync=True):
        """
        Share a file or a folder

        :param str path: The path of the file or folder to share
        :param list[cterasdk.core.types.ShareRecipient] recipients: A list of share recipients
        :param bool,optional as_project: Share as a team project, defaults to True when the item is a cloud folder else False
        :param bool,optional allow_reshare: Allow recipients to re-share this item, defaults to True
        :param bool,optional allow_sync: Allow recipients to sync this item, defaults to True when the item is a cloud folder else False
        :return: A list of all recipients added to the collaboration share
        :rtype: list[cterasdk.core.types.ShareRecipient]
        """
        return shares.share(self._core, self.get_object_path(path), recipients, as_project, allow_reshare, allow_sync)

    def add_share_recipients(self, path, recipients):
        """
        Add share recipients

        :param str path: The path of the file or folder
        :param list[cterasdk.core.types.ShareRecipient] recipients: A list of share recipients
        :return: A list of all recipients added
        :rtype: list[cterasdk.core.types.ShareRecipient]
        """
        return shares.add_share_recipients(self._core, self.get_object_path(path), recipients)

    def remove_share_recipients(self, path, accounts):
        """
        Remove share recipients

        :param str path: The path of the file or folder
        :param list[cterasdk.core.types.PortalAccount] accounts: A list of portal user or group accounts
        :return: A list of all share recipients removed
        :rtype: list[cterasdk.core.types.PortalAccount]
        """
        return shares.remove_share_recipients(self._core, self.get_object_path(path), accounts)

    def unshare(self, path):
        """
        Unshare a file or a folder
        """
        return shares.unshare(self._core, self.get_object_path(path))


class Backups(FileBrowser):

    def device_config(self, device, destination=None):
        """
        Download a device configuration file

        :param str device: The device name
        :param str,optional destination:
         File destination, if it is a directory, the original filename will be kept, defaults to the default directory
        """
        try:
            destination = destination if destination is not None else f'{cterasdk.settings.downloads.location}/{device}.xml'
            return self.download(f'backups/{device}/Device Configuration/db.xml', destination)
        except CTERAException as error:
            logging.getLogger().error('Failed downloading configuration file. %s', {'device': device, 'error': error.response.reason})
            raise error

    @property
    def base(self):
        return f'{super().base}/backups'



from .core import GlobalAdmin, ServicesPortal  # noqa: E402, F401
from .edge import Edge  # noqa: E402, F401
from .drive import Drive  # noqa: E402, F401
from .data import DataServices  # noqa: E402, F401



from datetime import datetime
import logging
import cterasdk.settings

from ..lib import FileSystem
from ..exceptions import InputError
from .base_command import BaseCommand


class DebugLevel:
    none = "none"
    error = "error"
    warning = "warning"
    info = "info"
    debug = "debug"
    process = "process"
    samba = "samba"
    aapi = "aapi"
    storage = "storage"
    upload = "upload"
    http = "http"
    dns = "dns"
    ntp = "ntp"
    backup = "backup"
    db = "db"
    files = "files"
    cttp = "cttp"
    cttp_data = "cttp_data"
    rsync = "rsync"
    alert = "alert"
    cbck = "cbck"
    av = "av"
    index = "index"
    auth = "auth"
    license = "license"
    collaboration = "collaboration"
    error_abort = "error_abort"
    cloud_extender = "cloud_extender"
    apps = "apps"
    evictor = "evictor"
    evictor_verbose = "evictor_verbose"
    caching = "caching"


class Support(BaseCommand):
    """ Edge Filer Support APIs """

    def set_debug_level(self, *levels):
        """
        Set the debug level
        """
        options = [v for k, v in DebugLevel.__dict__.items() if not k.startswith('_')]
        cli_command = 'dbg level'
        for level in levels:
            if level not in options:
                raise InputError('Invalid debug level', level, options)
            cli_command = cli_command + ' ' + level
        return self._edge.cli.run_command(cli_command)

    def get_support_report(self):
        """ Download support report """
        filename = 'Support-' + self._edge.host() + datetime.now().strftime('_%Y-%m-%dT%H_%M_%S') + '.zip'
        logging.getLogger().info('Downloading support report. %s', {'host': self._edge.host()})
        handle = self._edge.api.handle('/supportreport')
        filepath = FileSystem.instance().save(cterasdk.settings.downloads.location, filename, handle)
        logging.getLogger().info('Support report downloaded. %s', {'filepath': filepath})



from ..clients.synchronous import clients
from .services import Management
from .endpoints import EndpointBuilder

from ..edge import backup
from ..edge import cli
from ..edge import logs
from ..edge import services
from ..edge import session
from ..edge import support
from ..edge import sync


class Clients:

    def __init__(self, drive, Portal):
        if Portal:
            drive._Portal = Portal
            drive._generic.shutdown()
            drive._ctera_session.start_remote_session(Portal.session())
            self._api = clients.API(EndpointBuilder.new(drive.base), Portal._generic._async_session, lambda *_: True)
        else:
            self._api = clients.API(EndpointBuilder.new(drive.base, '/admingui/api'), drive._generic._async_session, drive._authenticator)


class Drive(Management):

    def __init__(self, host=None, port=None, https=True, Portal=None, *, base=None):
        super().__init__(host, port, https, base=base)
        self._ctera_session = session.Session(self.host())

        self._initialize(Portal)

        self.backup = backup.Backup(self)
        self.cli = cli.CLI(self)
        self.logs = logs.Logs(self)
        self.services = services.Services(self)
        self.support = support.Support(self)
        self.sync = sync.Sync(self)

    def _initialize(self, Portal):
        if Portal:
            self._generic.shutdown()
            async_session = Portal._generic._async_session  # pylint: disable=protected-access
            self._ctera_session.start_remote_session(Portal.session())
            self._api = clients.API(EndpointBuilder.new(self.base), async_session, lambda *_: True)
        else:
            async_session = self._generic._async_session  # pylint: disable=protected-access
            self._api = clients.API(EndpointBuilder.new(self.base, '/admingui/api'), async_session, self._authenticator)

    @property
    def _session_id_key(self):
        return '_cteraSessionId_'

    @property
    def _login_object(self):
        raise NotImplementedError("Logins to the 'Drive App' are not enabled.")



class Registry:  # pylint: disable=unused-private-member

    __instance = None

    @staticmethod
    def instance():
        if Registry.__instance is None:
            Registry()
        return Registry.__instance

    def __init__(self):
        if Registry.__instance is not None:
            raise Exception("Registry is a singleton class.")
        self.registry = {}
        Registry.__instance = self

    def register(self, key, value):
        self.registry[key] = value

    def get(self, key):
        return self.registry.get(key)

    def remove(self, key):
        return self.registry.pop(key, None)



import logging

from ..common import Object
from .types import DeduplicationStatus
from .base_command import BaseCommand


class Dedup(BaseCommand):
    """ Edge Filer Local Deduplication APIs """

    def __init__(self, edge):
        super().__init__(edge)
        self.regen = Regeneration(self._edge)

    def enable(self, reboot=False, wait=False):
        """
        Enable local deduplication

        :param bool reboot: Reboot, defaults to ``False``
        :param bool,optional wait: Wait for reboot to complete, defaults to False
        """
        logging.getLogger().info("Enabling local deduplication.")
        response = self._edge.api.put('/config/dedup/useLocalMapFileDedup', True)
        self._wait_for_reboot(reboot, wait)
        return response

    def disable(self, reboot=False, wait=False):
        """
        Disable local deduplication

        :param bool reboot: Reboot, defaults to ``False``
        :param bool,optional wait: Wait for reboot to complete, defaults to False
        """
        logging.getLogger().info("Disabling local deduplication.")
        response = self._edge.api.put('/config/dedup/useLocalMapFileDedup', False)
        self._wait_for_reboot(reboot, wait)
        return response

    def status(self):
        """
        Get the de-duplication status

        :returns: An object including the deduplication status
        :rtype: cterasdk.edge.types.DeduplicationStatus
        """
        size = self._edge.api.execute('/config/cloudsync/cloudExtender', 'allFilesTotalUsedBytes')
        usage = self._edge.api.execute('/config/cloudsync/cloudExtender', 'storageUsedBytes')
        return DeduplicationStatus(size, usage)

    def _wait_for_reboot(self, reboot, wait):
        if reboot:
            self._edge.power.reboot(wait)


class Regeneration(BaseCommand):
    """ Edge Filer Local Deduplication Regeneration APIs """

    def run(self):
        """
        Run the regeneration process
        """
        logging.getLogger().info("Executing the dedup regeneration process.")
        return self._edge.api.execute('/config/dedup', 'regenerate', Object())

    def status(self):
        """
        Get the regeneration process statistics
        """
        return self._edge.api.get('/proc/dedup/regenerate/general')



import re
from datetime import datetime
from collections import namedtuple

from .object import Object
from .utils import df_military_time, day_of_week
from .enum import FileCriteria, BooleanFunction, Application, ScheduleType


class PolicyRule:

    def __init__(self, assignment, criteria):
        self.assignment = assignment
        self.criteria = criteria


class PolicyRuleConverter:

    @staticmethod
    def convert(rule, classname, property_name, assignment=None):
        param = Object()
        param._classname = classname  # pylint: disable=protected-access
        setattr(param, property_name, assignment if assignment else rule.assignment)
        param.filterRule = rule.criteria
        return param


class Operator(Object):

    def __init__(self, right):
        self._classname = self.__class__.__name__  # pylint: disable=protected-access
        self.right = right


class IsOperator(Operator):
    pass


class BeginsWithOperator(Operator):
    pass


class EndsWithOperator(Operator):
    pass


class ContainsOperator(Operator):
    pass


class IsOneOfOperator(Operator):
    pass


class LessThanOperator(Operator):
    pass


class MoreThanOperator(Operator):
    pass


class BeforeOperator(Operator):
    pass


class AfterOperator(Operator):
    pass


class AdvancedFilterRule(Object):

    def __init__(self, classname, field, operator):
        self._classname = self.__class__.__name__  # pylint: disable=protected-access
        self.className = classname
        self.fieldName = field
        self.operator = operator


class CriteriaBuilder:

    def __init__(self, criteria_type, criteria_field):
        self.criteria_type = criteria_type
        self.criteria_field = criteria_field
        self.operator = None

    def build(self):
        return AdvancedFilterRule(self.criteria_type, self.criteria_field, self.operator)


class ListCriteriaBuilder(CriteriaBuilder):

    def include(self, values):
        self.operator = IsOneOfOperator(values)
        return self


class IntegerCriteriaBuilder(CriteriaBuilder):

    def less_than(self, value):
        self.operator = LessThanOperator(value)
        return self

    def more_than(self, value):
        self.operator = MoreThanOperator(value)
        return self


class DateTimeCriteriaBuilder(CriteriaBuilder):

    def before(self, value):
        self.operator = BeforeOperator(value)
        return self

    def after(self, value):
        self.operator = AfterOperator(value)
        return self


class StringCriteriaBuilder(CriteriaBuilder):

    def equals(self, value):
        self.operator = IsOperator(value)
        return self

    def startswith(self, value):
        self.operator = BeginsWithOperator(value)
        return self

    def endswith(self, value):
        self.operator = EndsWithOperator(value)
        return self

    def contains(self, value):
        self.operator = ContainsOperator(value)
        return self

    def isoneof(self, values):
        self.operator = IsOneOfOperator(values)
        return self


class ThrottlingRule:
    """
    Throttling Rule

    :ivar int upload: Throttling rate upstream (Kilobits)
    :ivar int download: Throttling rate downstream (Kilobits)
    :ivar str start: Start time
    :ivar str end: End time
    :ivar list[str] days: Days
    """

    def __init__(self):
        self.upload = None
        self.download = None
        self.schedule = None

    def to_server_object(self):
        param = Object()
        param._classname = 'SyncThrottlingSettings'  # pylint: disable=protected-access
        param.inKbitsPerSecond = self.download
        param.outKbitsPerSecond = self.upload
        param.schedule = self.schedule
        return param

    @staticmethod
    def from_server_object(param):
        r = ThrottlingRule()
        r.download = param.inKbitsPerSecond
        r.upload = param.outKbitsPerSecond
        r.schedule = param.schedule
        return r

    def __str__(self):
        return str(dict(upload_kbps=self.upload, download_kbps=self.download,
                        start=self.schedule.start, end=self.schedule.end, days=[day_of_week(day) for day in self.schedule.days]))


class ThrottlingRuleBuilder:
    """
    Bandwidth Throttling Rule Builder
    """

    def __init__(self):
        self.param = ThrottlingRule()

    def upload(self, kbps):
        """
        Throttle bandwidth upstream

        :param int kbps: Kilobits per second
        """
        self.param.upload = kbps
        return self

    def download(self, kbps):
        """
        Throttle bandwidth downstream

        :param int kbps: Kilobits per second
        """
        self.param.download = kbps
        return self

    def schedule(self, schedule):
        """
        Set the throttling rule schedule

        :param cterasdk.common.types.TimeRange schedule: Schedule
        """
        self.param.schedule = schedule
        return self

    def build(self):
        """
        Build the throttling rule
        """
        errors = [k for k, v in self.param.__dict__.items() if v is None]
        if errors:
            raise ValueError(f'No value for required field: {errors}')
        return self.param


class FileFilterBuilder:

    Type = 'File'

    @staticmethod
    def extensions():
        """Filter files by extension"""
        return ListCriteriaBuilder(FileFilterBuilder.Type, FileCriteria.Type)

    @staticmethod
    def names():
        """Filter files by names"""
        return ListCriteriaBuilder(FileFilterBuilder.Type, FileCriteria.Name)

    @staticmethod
    def name():
        """Filter files by name pattern"""
        return StringCriteriaBuilder(FileFilterBuilder.Type, FileCriteria.Name)

    @staticmethod
    def paths():
        """Filter files by path"""
        return ListCriteriaBuilder(FileFilterBuilder.Type, FileCriteria.Path)

    @staticmethod
    def path():
        """Filter files by path pattern"""
        return StringCriteriaBuilder(FileFilterBuilder.Type, FileCriteria.Path)

    @staticmethod
    def size():
        """Filter files by size"""
        return IntegerCriteriaBuilder(FileFilterBuilder.Type, FileCriteria.Size)

    @staticmethod
    def last_modified():
        """Filter files by last modification date"""
        return DateTimeCriteriaBuilder(FileFilterBuilder.Type, FileCriteria.Modified)


class DirectoryEntryFactory:

    @staticmethod
    def root(included):
        return DirEntry('root', included=included)


class FileEntry(Object):

    def __init__(self, name, display_name=None, included=None):
        self._classname = self.__class__.__name__  # pylint: disable=protected-access
        self.name = name
        self.displayName = display_name
        self.isIncluded = included


class DirEntry(FileEntry):

    def __init__(self, name, display_name=None, included=None, children=None):
        super().__init__(name, display_name, included)
        self.children = children


class BackupSet(Object):

    def __init__(self, name, directory_tree=None, filter_rules=None, defaults_dirs=None,
                 template_dirs=None, enabled=True, boolean_function=None, comment=None):
        self._classname = self.__class__.__name__  # pylint: disable=protected-access
        self.name = name
        self.isEnabled = enabled
        self.directoryTree = directory_tree if directory_tree else DirectoryEntryFactory.root(True)
        self.booleanFunction = boolean_function if boolean_function else BooleanFunction.AND
        self.templateDirectories = template_dirs
        self.defaultDirs = defaults_dirs
        self.comment = comment
        self.filterRules = filter_rules


class FilterBackupSet(BackupSet):
    pass


class ApplicationBackupSet(BackupSet):

    def __init__(self, apps):

        comment = 'Backup all the specified applications'
        name = 'Applications'

        if apps == Application.All:
            super().__init__(name=name, directory_tree=DirEntry(name, included=True), comment=comment)
        else:
            directory_tree = DirEntry(name, included=False, children=[DirEntry(app, included=True) for app in apps])
            super().__init__(name=name, directory_tree=directory_tree, comment=comment)

        self._classname = None


class TaskSchedule(Object):

    def __init__(self):
        self._classname = 'TaskSchedule'  # pylint: disable=protected-access
        self.mode = None


class BackupScheduleBuilder:

    @staticmethod
    def interval(hours=None, minutes=None):
        """
        Schedule backup to periodically, defaults to 24 hours

        :param int hours: Hours
        :param int minutes: Minutes
        """
        param = TaskSchedule()
        param.mode = ScheduleType.Interval
        param.interval = Object()  # pylint: disable=attribute-defined-outside-init
        param.interval.hours = hours if hours is not None else 24
        param.interval.minutes = minutes if minutes is not None else 0
        return param

    @staticmethod
    def window(time_range):
        """
        Schedule backup to run at a specific time

        :param cterasdk.common.types.TimeRange time_range: Time range
        """
        param = TaskSchedule()
        param.mode = ScheduleType.Window
        param.window = time_range  # pylint: disable=attribute-defined-outside-init
        return param


class TimeRange:
    """ Class representing a task schedule """

    def __init__(self):
        self.param = Object()
        self.param._classname = 'TimeRange'  # pylint: disable=protected-access
        self.param.start = None
        self.param.end = None
        self.param.days = None
        self.param.termOnEnd = False

    def start(self, start):
        """
        Start time

        :param str start: A military time string 'hh:mm:ss' or a datetime object
        """
        self.param.start = TimeRange._infer_time_range(start)
        return self

    def end(self, end):
        """
        End time

        :param str end: A military time string 'hh:mm:ss' or a datetime object
        """
        self.param.end = TimeRange._infer_time_range(end)
        return self

    def days(self, days):
        """
        Set days

        :param list[cterasdk.common.enum.DayOfWeek] days: A list of days
        """
        self.param.days = days
        return self

    def terminate_at_endtime(self):
        """
        Terminate at end time, defaults to teminate on completion.
        """
        self.param.termOnEnd = True
        return self

    def build(self):
        """
        Build the time range
        """
        if self.param.termOnEnd and self.param.end is None:
            raise ValueError('End time required')
        if self.param.start is None:
            raise ValueError('Start time required')
        return self.param

    @staticmethod
    def _infer_time_range(time):
        if isinstance(time, datetime):
            return df_military_time(time)
        if isinstance(time, str):
            match = re.search('^[012][0-9]:[0-5][0-9]:[0-5][0-9]$', time)
            if match:
                return match.group(0)
            raise ValueError("Invalid time format. Expected 'hh:mm:ss'")
        raise ValueError("Invalid format. Expected 'datetime' or 'str'")

    def __str__(self):
        return str(self.param)


class ADDomainIDMapping(Object):
    """
    Base Class for Directory Service ID Mapping

    :ivar str domain: The domain flat name
    :param int start: The minimum id to use for mapping
    :param int end: The maximum id to use for mapping
    """
    def __init__(self, domain, start, end):
        self._classname = 'ADDomainIDMapping'
        self.domainFlatName = domain
        self.minID = start
        self.maxID = end


class SoftwareUpdatePolicyBuilder:
    """
    Software Update Policy Builder
    """

    def __init__(self):
        self.param = SoftwareUpdatesTopic(enabled=True, reboot_after_update=True, reboot_when=None)

    def download_and_install(self, download_and_install):
        """
        Download and install updates automatically.
        :param bool download_and_install: Set ``True`` to enable and ``False`` to disable.
        """
        self.param.download_and_install(download_and_install)
        return self

    def reboot_after_update(self, reboot_after_update):
        """
        Restart device automatically after installing new firmware

        :param bool reboot_after_update: Set ``True`` to enable and ``False`` to disable.
        """
        self.param.reboot_after_update(reboot_after_update)
        return self

    def schedule(self, schedule):
        """
        Set the throttling rule schedule

        :param cterasdk.common.types.TimeRange schedule: Schedule
        """
        self.param.schedule(schedule)
        return self

    def build(self):
        """
        Build the Software Update Policy
        """
        if self.param.rebootAfterUpdate is False:
            self.param.reboot_asap()
        return self.param


class SoftwareUpdatesTopic(Object):

    def __init__(self, enabled, reboot_after_update, reboot_when):
        self._classname = "SoftwareUpdatesSettings"
        self.enabled = enabled if enabled else None
        self.rebootAfterUpdate = reboot_after_update if reboot_after_update else None
        self.rebootWhen = reboot_when if reboot_when else None

    def download_and_install(self, enabled):
        self.enabled = enabled

    def reboot_after_update(self, enabled):
        self.rebootAfterUpdate = enabled

    def reboot_asap(self):
        self.rebootWhen = None

    def schedule(self, window):
        self.rebootWhen = Object()
        self.rebootWhen.mode = ScheduleType.Window
        self.rebootWhen.window = window


ConsentPage = namedtuple('ConsentPage', ('header', 'body'))
ConsentPage.__doc__ = 'Tuple holding the consent page header and body'
ConsentPage.header.__doc__ = 'The consent page header'
ConsentPage.body.__doc__ = 'The consent page body'



from .base_command import BaseCommand


class Domains(BaseCommand):
    """
    Portal Domains Management APIs
    """

    def list_domains(self):
        """
        List all domains

        :return list: List of all domains
        """
        return self._core.api.get('/domains')



from pathlib import Path
import yaml
from .convert import tojsonstr, fromjsonstr


sessions, downloads = None, None


def default_settings():
    global sessions, downloads  # # pylint: disable=global-statement
    with open(Path(__file__).parent.absolute().joinpath('settings.yml'), 'r', encoding='utf-8') as f:
        sdk_settings = yaml.safe_load(f)
        sessions = convert_to_object(sdk_settings['sessions'])
        downloads = convert_to_object(sdk_settings['downloads'])


def convert_to_object(data):
    return fromjsonstr(tojsonstr(data))


default_settings()



import logging

from .base_command import BaseCommand
from ..lib import X509Certificate, PrivateKey, create_certificate_chain
from ..common import Object


class SSL(BaseCommand):
    """ Edge Filer SSL APIs """

    def enable_http(self):
        """
        Enable HTTP access
        """
        self._set_force_https(False)

    def disable_http(self):
        """
        Disable HTTP access
        """
        self._set_force_https(True)

    def is_http_disabled(self):
        """
        Check if HTTP access is disabled
        """
        return self._get_force_https()

    def is_http_enabled(self):
        """
        Check if HTTP access is enabled
        """
        return not self._get_force_https()

    def _get_force_https(self):
        return self._edge.api.get('/config/fileservices/webdav/forceHttps')

    def _set_force_https(self, force):
        self._edge.api.put('/config/fileservices/webdav/forceHttps', force)

    def remove_storage_ca(self):
        """
        Remove object storage trusted CA certificate
        """
        self._edge.api.put('/config/extStorageTrustedCA', None)

    def get_storage_ca(self):
        """
        Get object storage trusted CA certificate
        """
        return self._edge.api.get('/status/extStorageTrustedCA')

    def import_storage_ca(self, certificate):
        """
        Import the object storage trusted CA certificate

        :param str certificate: The PEM-encoded certificate or a path to the PEM-encoded server certificate file
        """
        logging.getLogger().info('Setting trusted object storage CA certificate')
        param = Object()
        param._classname = 'ExtTrustedCA'  # pylint: disable=protected-access
        param.certificate = X509Certificate.load_certificate(certificate).pem_data.decode('utf-8')
        logging.getLogger().info("Uploading object storage certificate.")
        response = self._edge.api.put('/config/extStorageTrustedCA', param)
        logging.getLogger().info("Uploaded object storage certificate.")
        return response

    def import_certificate(self, private_key, *certificates):
        """
        Import the Edge Filer's web server's SSL certificate

        :param str private_key: The PEM-encoded private key, or a path to the PEM-encoded private key file
        :param list[str] certificates: The PEM-encoded certificates, or a list of paths to the PEM-encoded certificates
        """

        key_object = PrivateKey.load_private_key(private_key)
        certificates = [X509Certificate.load_certificate(certificate) for certificate in certificates]
        certificate_chain = [certificate.pem_data.decode('utf-8') for certificate in create_certificate_chain(*certificates)]
        server_certificate = ''.join([key_object.pem_data.decode('utf-8')] + certificate_chain)
        logging.getLogger().info("Uploading SSL certificate.")
        response = self._edge.api.put('/config/certificate', f"\n{server_certificate}")
        logging.getLogger().info("Uploaded SSL certificate.")
        return response



import logging

from .base_command import BaseCommand
from ..common import Object
from ..core.enum import Severity, Mode, IPProtocol
from ..exceptions import CTERAException


class Syslog(BaseCommand):
    """
    Portal Syslog Management APIs
    """

    # TODO: upload_ca_certificate  # pylint: disable=W0511
    # TODO: upload_client_certificate  # pylint: disable=W0511

    def is_enabled(self):
        """
        Check if forwarding log messages over syslog is enabled
        """
        return self._core.api.get('/settings/logsSettings/syslogConfig/mode') == Mode.Enabled

    def get_configuration(self):
        """
        Retrieve the syslog server configuration
        """
        return self._core.api.get('/settings/logsSettings/syslogConfig')

    def enable(self, server, port=514, protocol=IPProtocol.UDP, min_severity=Severity.INFO):
        """
        Enable Syslog

        :param str server: Syslog server address
        :param int,optional port: Syslog server port
        :param cterasdk.core.enum.IPProtocol,optional protocol: Syslog server IP protocol
        :param cterasdk.core.enum.Severity,optional min_severity: Minimum log severity to forward
        """
        param = Object()
        param._classname = 'PortalSyslogConfig'  # pylint: disable=protected-access
        param.mode = Mode.Enabled
        param.server = server
        param.minSeverity = min_severity
        param.port = port
        param.protocol = protocol
        param.useClientCertificate = False
        logging.getLogger().info('Enabling syslog.')
        response = self._core.api.put('/settings/logsSettings/syslogConfig', param)
        logging.getLogger().info('Syslog enabled.')
        return response

    def modify(self, server=None, port=None, protocol=None, min_severity=None):
        """
        Modify Syslog log forwarding configuration

        :param str server: Syslog server address
        :param int,optional port: Syslog server port
        :param cterasdk.core.enum.IPProtocol,optional protocol: Syslog server IP protocol
        :param cterasdk.core.enum.Severity,optional min_severity: Minimum log severity to forward
        """
        current_config = self.get_configuration()
        if current_config.mode == Mode.Disabled:
            raise CTERAException("Syslog configuration cannot be modified when disabled")
        if server:
            current_config.server = server
        if port:
            current_config.port = port
        if protocol:
            current_config.protocol = protocol
        if min_severity:
            current_config.minSeverity = min_severity

        logging.getLogger().info("Updating syslog server configuration.")
        self._core.api.put('/settings/logsSettings/syslogConfig', current_config)
        logging.getLogger().info(
            "Syslog server configured. %s",
            {
                'server': current_config.server,
                'port': current_config.port,
                'protocol': current_config.protocol,
                'minSeverity': current_config.minSeverity
            }
        )

    def disable(self):
        logging.getLogger().info('Disabling syslog.')
        response = self._core.api.put('/settings/logsSettings/syslogConfig/mode', Mode.Disabled)
        logging.getLogger().info('Syslog disabled.')
        return response



from ..clients.synchronous import clients
from ..objects.endpoints import EndpointBuilder


class IO:

    def __init__(self, core):
        async_session = self._generic._async_session
        self._folders = clients.Folders(EndpointBuilder.new(core.base, core.context, '/folders/folders'),
                                        async_session, core._authenticator)
        self._upload = clients.Upload(EndpointBuilder.new(core.base, core.context, '/upload/folders'), async_session, core._authenticator)
        self._webdav = clients.Dav(EndpointBuilder.new(core.base, core.context, '/webdav'), async_session, core._authenticator)



import logging
from . import common


def listdir(core, path, depth=None, include_deleted=False):
    depth = depth if depth is not None else 1
    builder = common.FetchResourcesParamBuilder().root(path.encoded_fullpath()).depth(depth)
    if include_deleted:
        builder.include_deleted()
    param = builder.build()
    if depth > 0:
        return common.objects_iterator(core, param)
    return common.fetch_resources(core, param)


def walk(core, base, path, include_deleted=False):
    paths = [common.get_object_path(base, path)]
    while len(paths) > 0:
        path = paths.pop(0)
        elements = listdir(core, path, include_deleted=include_deleted)
        for element in elements:
            if element.isFolder:
                paths.append(common.get_object_path(base, element))
            yield element


def mkdir(core, path):
    param = common.get_create_dir_param(path.name(), path.parent().encoded_fullpath())
    logging.getLogger().info('Creating directory. %s', {'path': str(path.relative)})
    response = core.api.execute('', 'makeCollection', param)
    common.raise_for_status(response, str(path.relative))
    logging.getLogger().info('Directory created. %s', {'path': str(path.relative)})


def makedirs(core, path):
    directories = path.parts()
    for i in range(1, len(directories) + 1):
        path = common.get_object_path(path.base, '/'.join(directories[:i]))
        try:
            mkdir(core, path)
        except common.ItemExists:
            pass


def rename(core, path, name):
    param = common.ActionResourcesParam.instance()
    logging.getLogger().info('Renaming item. %s', {'path': str(path.relative), 'name': name})
    param.add(common.SrcDstParam.instance(src=path.fullpath(), dest=path.parent().joinpath(name).fullpath()))
    return core.api.execute('', 'moveResources', param)


def remove(core, *paths):
    param = common.ActionResourcesParam.instance()
    paths = [paths] if not isinstance(paths, tuple) else paths
    for path in paths:
        logging.getLogger().info('Deleting item. %s', {'path': str(path.relative)})
        param.add(common.SrcDstParam.instance(src=path.fullpath()))
    return core.api.execute('', 'deleteResources', param)


def recover(core, *paths):
    param = common.ActionResourcesParam.instance()
    paths = [paths] if not isinstance(paths, tuple) else paths
    for path in paths:
        logging.getLogger().info('Recovering item. %s', {'path': str(path.relative)})
        param.add(common.SrcDstParam.instance(src=path.fullpath()))
    return core.api.execute('', 'restoreResources', param)


def copy(core, *paths, destination=None):
    param = common.ActionResourcesParam.instance()
    paths = [paths] if not isinstance(paths, tuple) else paths
    for path in paths:
        logging.getLogger().info('Copying item. %s', {'path': str(path.relative), 'to': str(destination.relative)})
        param.add(common.SrcDstParam.instance(src=path.fullpath(), dest=destination.joinpath(path.name()).fullpath()))
    return core.api.execute('', 'copyResources', param)


def move(core, *paths, destination=None):
    param = common.ActionResourcesParam.instance()
    paths = [paths] if not isinstance(paths, tuple) else paths
    for path in paths:
        logging.getLogger().info('Copying item. %s', {'path': str(path.relative), 'to': str(destination.relative)})
        param.add(common.SrcDstParam.instance(src=path.fullpath(), dest=destination.joinpath(path.name()).fullpath()))
    return core.api.execute('', 'moveResources', param)



from . import uri


class EndpointBuilder:

    def __init__(self, base, *segments):
        self._base = base
        self._segments = segments

    def __call__(self, *segments):
        return uri.create(self._base, *(self._segments + segments))

    @staticmethod
    def new(base, *segments):
        return EndpointBuilder(base, *segments)



import logging
from .base_command import BaseCommand


class AIO(BaseCommand):
    """
    Edge Filer AIO APIs
    """
    def is_enabled(self):
        """
        Is AIO enabled

        :return: True is AIO is enabled, else False
        :rtype: bool
        """
        cifs = self._edge.api.get('/config/fileservices/cifs')
        return cifs.robustMutexes and (cifs.aioReadThreshold > 0) and (cifs.aioWriteThreshold > 0)

    def enable(self):
        """
        Enable AIO
        """
        logging.getLogger().info('Enabling asynchronous io.')
        self._async_io(True, 1, 1)
        logging.getLogger().info('Asynchronous io enabled.')

    def disable(self):
        """
        Disable AIO
        """
        logging.getLogger().info('Disabling asynchronous io.')
        self._async_io(False, 0, 0)
        logging.getLogger().info('Asynchronous io disabled.')

    def _async_io(self, robustMutexes, aioReadThreshold, aioWriteThreshold):
        logging.getLogger().debug('Obtaining CIFS server settings.')

        cifs = self._edge.api.get('/config/fileservices/cifs')
        cifs.robustMutexes = robustMutexes
        cifs.aioReadThreshold = aioReadThreshold
        cifs.aioWriteThreshold = aioWriteThreshold

        logging.getLogger().debug('Updating CIFS server settings.')

        self._edge.api.put('/config/fileservices/cifs', cifs)



import logging
from pathlib import PurePosixPath
from ...exceptions import CTERAException
from ...objects import uri


class Path:

    def __init__(self, relative, base):
        self._base = PurePosixPath(base)
        self._relative = PurePosixPath(relative)
        if self._base.joinpath(self._relative) == self._relative:
            raise ValueError('You must specify a relative path. Omit leading / characters')

    @property
    def base(self):
        return self._base

    @property
    def relative(self):
        return self._relative

    def name(self):
        return self.relative.name

    def parent(self):
        return Path(str(self.relative.parent), str(self.base))

    def fullpath(self):
        return str(self.base.joinpath(self.relative))

    def encoded_fullpath(self):
        return uri.quote(self.fullpath())

    def encoded_parent(self):
        return uri.quote(str(self.parent()))

    def joinpath(self, path):
        return Path(str(self.relative.joinpath(path)), str(self.base))

    def parts(self):
        return self.relative.parts

    def __str__(self):
        return self.fullpath()


class ItemExists(CTERAException):
    """Already exists"""


class Forbidden(CTERAException):
    """Forbidden"""


def raise_for_status(response, path):
    error = {
        "File exists": ItemExists(),
        "Creating a folder in this location is forbidden": Forbidden(),
    }.get(response, None)
    try:
        if error:
            raise error
    except ItemExists as error:
        logging.getLogger().warning('A file or folder with the same name already exists. %s', {'path': path})
        raise error
    except Forbidden as error:
        logging.getLogger().error('Creating a folder in this location is forbidden. %s', {'name': path})
        raise error


def get_object_path(base, relative):
    return Path(relative, base)



import re
import logging

from ..common import parse_base_object_ref
from ..exceptions import CTERAException


def remote_access(device, Portal):
    device_tenant = parse_base_object_ref(device.portal).name
    device_name = device.name
    logging.getLogger().info("Enabling remote access. %s", {'tenant': device_tenant, 'device': device_name})
    token = authn_token(Portal, device_tenant, device_name)
    device_object = create_device_object(device)
    device_object.sso(token)
    logging.getLogger().info("Enabled remote access. %s", {'tenant': device_tenant, 'device': device_name})
    return device_object


def create_device_object(device):
    device_object = device.__class__(base=re.sub(r'^http(?=:)', 'https', device.remoteAccessUrl))
    return device_object


def authn_token(Portal, device_tenant, device_name):
    logging.getLogger().debug("Retrieving SSO Ticket. %s", {'tenant': device_tenant, 'device': device_name})
    token = Portal.api.execute(f"/portals/{device_tenant}/devices/{device_name}", 'singleSignOn')
    if not token:
        logging.getLogger().error('Failed to Retrieve SSO Ticket. %s', {'tenant': device_tenant, 'device': device_name})
        raise CTERAException('Failed to Retrieve SSO Ticket.')
    logging.getLogger().debug("Retrieved SSO Ticket. %s", {'tenant': device_tenant, 'device': device_name})
    return token



from .services import CTERA
from .endpoints import EndpointBuilder
from ..clients.asynchronous import clients


from ..asynchronous.core import login
from ..asynchronous.core import cloudfs
from ..asynchronous.core import notifications
from ..asynchronous.core import users


class Clients:

    def __init__(self, services):
        self.v1 = V1(services)
        self.v2 = V2(services)


class V1:

    def __init__(self, services):
        session = services._generic._async_session
        self.api = clients.AsyncAPI(EndpointBuilder.new(services.base, services.context, '/api'), session, services._authenticator)


class V2:

    def __init__(self, services):
        session = services._generic._async_session
        self.api = clients.AsyncJSON(EndpointBuilder.new(services.base, services.context, '/v2/api'), session, services._authenticator)


class DataServices(CTERA):

    async def __aenter__(self):
        return self

    def __init__(self, host, port=None, https=True):
        super().__init__(host, port, https, base=None)
        self._generic = clients.AsyncClient(EndpointBuilder.new(self.base), authenticator=self._authenticator)
        self._ctera_clients = Clients(self)

        self.cloudfs = cloudfs.CloudFS(self)
        self.notifications = notifications.Notifications(self)
        self.users = users.Users(self)

    @property
    def v1(self):
        return self.clients.v1

    @property
    def v2(self):
        return self.clients.v2

    async def login(self, username, password):
        return await self._login_object.login(username, password)

    async def logout(self):
        await self._login_object.logout()
        await self.clients.v1.api.shutdown()

    @property
    def _login_object(self):
        return login.Login(self)

    @property
    def context(self):
        return 'admin'

    def _authenticator(self, url):
        return True

    async def __aexit__(self, exc_type, exc, tb):
        await self._generic.shutdown()



import logging

from .base_command import BaseCommand
from ..common import Object
from .enum import TaskType, SourceType


class CTERAMigrate(BaseCommand):
    """Edge Filer Migration Tool APIs """

    def __init__(self, edge):
        super().__init__(edge)
        self.discovery = Discovery(self)
        self.migration = Migration(self)

    def login(self):
        """
        Login to CTERA Migrate
        """
        return self._edge.migrate.login()  # pylint: disable=protected-access

    def list_shares(self, credentials):
        """
        Log in

        :param cterasdk.edge.types.HostCredentials credentials: Target host credentials
        """
        param = Object()
        param.host = credentials.host
        param.user = credentials.username
        setattr(param, 'pass', credentials.password)
        return [share.name for share in
                self._edge.migrate.post('/inventory/shares', param).shares]  # pylint: disable=protected-access

    def list_tasks(self, deleted=False):
        """
        List tasks

        :param bool,optional deleted: List deleted tasks, defaults to ``False``
        :returns: List of tasks
        :rtype: list(cterasdk.common.object.Object)
        """
        tasks = self._edge.migrate.get('/tasks/list', params={'deleted': int(deleted)}).tasks  # pylint: disable=W0212
        return [Task.from_server_object(task) for task in tasks.__dict__.values()] if tasks else []

    def delete(self, tasks):
        """
        Delete tasks

        :param list(cterasdk.common.object.Object) tasks: List of tasks
        """
        return self._delete_or_restore(tasks, 'delete')

    def restore(self, tasks):
        """
        Recover tasks

        :param list(cterasdk.common.object.Object) tasks: List of tasks
        """
        return self._delete_or_restore(tasks, 'restore')

    def _delete_or_restore(self, tasks, action):
        param = Object()
        param.task_ids = [task.id for task in tasks]
        return self._edge.migrate.post(f'/tasks/{action}', param)  # pylint: disable=protected-access

    def start(self, task):
        """
        Start a task

        :param cterasdk.common.object.Object task: Task object
        """
        return self._update_task(task, 'enable')

    def stop(self, task):
        """
        Stop a task

        :param cterasdk.common.object.Object task: Task object
        """
        return self._update_task(task, 'disable')

    def _update_task(self, task, action):
        """
        Enable or disable a task
        """
        param = Object()
        param.task_id = task.id
        return self._edge.migrate.post(f'/tasks/{action}', param)  # pylint: disable=protected-access

    def details(self, task):
        """
        Get task details
        """
        response = self._edge.migrate.get('/tasks/history', params={'id': task.id})  # pylint: disable=protected-access
        if response.history:
            return Jobs(response.history)
        logging.getLogger().error('Task not found. %s', {'task_id': task.id})
        return None

    def results(self, task):
        if task.type == 'discovery':
            return self._edge.migrate.get('/discovery/results',  # pylint: disable=protected-access
                                          params={'id': task.id}).discovery
        if task.type == 'migration':
            return self._edge.migrate.get('/migration/results',  # pylint: disable=protected-access
                                          params={'id': task.id}).migration
        logging.getLogger().error('Could not determine task type. %s', {'id': task.id, 'type': task.type, 'name': task.name})
        return None


class Jobs:
    """Class representing task jobs"""

    def __init__(self, jobs):
        self._all = jobs

    @property
    def latest(self):
        """
        Get the latest job of a task
        """
        if self._all:
            return self._all[0]
        return None

    @property
    def all(self):
        """
        Get all jobs of a task
        """
        return self._all


class TaskManager:
    """Class representing a migration tool task"""

    def __init__(self, ctera_migrate):
        self._ctera_migrate = ctera_migrate

    @staticmethod
    def _create_add_parameter(name, credentials, shares, auto_start=False, notes=None):
        param = Object()
        param.name = name
        param.host = credentials.host
        param.user = credentials.username
        setattr(param, 'pass', credentials.password)
        param.shares = []
        if credentials.host_type:
            param.host_type = credentials.host_type
        for share in shares:
            share_param = Object()
            share_param.src = share
            share_param.dst = share
            param.shares.append(share_param)
        param.start = auto_start
        param.notes = notes
        return param

    def _add(self, param):
        task = self._ctera_migrate._edge.migrate.post('/tasks/create', param)  # pylint: disable=protected-access
        return Task(task.task_id, int(task.type), task.name)


class Discovery(TaskManager):

    def list_tasks(self, deleted=False):
        return [task for task in self._ctera_migrate.list_tasks(deleted) if task.type == 'discovery']  # pylint: disable=protected-access

    def add(self, name, credentials, shares, auto_start=False, log_every_file=False, notes=None):
        """
        Create a discovery task

        :param str name: Task name
        :param cterasdk.edge.types.HostCredentials credentials: Target host credentials
        :param bool,optional auto_start: Start task after creation, defaults to ``False``
        :param bool,optional log_every_file: Log every file, defaults to ``False``
        :param str,optional notes: Task notes

        :returns: Task
        :rtype: cterasdk.common.object.Object
        """
        param = TaskManager._create_add_parameter(name, credentials, shares, auto_start, notes)
        param.type = TaskType.Discovery
        param.discovery_log_files = int(log_every_file)
        return self._add(param)

    def update(self, task, name=None, notes=None):
        """
        Update a discovery task

        :param str,optional name: Task name
        :param str,optional notes: Task notes
        """
        param = Object()
        param.task_id = task.id
        if name:
            param.name = name
        if notes:
            param.notes = notes
        return self._ctera_migrate._edge.migrate.post('/tasks/update', param)  # pylint: disable=protected-access


class Migration(TaskManager):

    def list_tasks(self, deleted=False):
        return [task for task in self._ctera_migrate.list_tasks(deleted) if task.type == 'migration']  # pylint: disable=protected-access

    def add(self, name, credentials, shares, auto_start=False, access_time=None,  # pylint: disable=too-many-arguments
            winacls=True, cloud_folder=None, create_cloud_folder_per_share=False,
            compute_checksum=False, exclude=None, include=None, notes=None):
        """
        Create a discovery task

        :param str name: Task name
        :param cterasdk.edge.types.HostCredentials credentials: Target host credentials
        :param bool,optional auto_start: Start task after creation, defaults to ``False``
        :param bool,optional access_time: Copy last access time, defaults to ``None``
        :param bool,optional winacls: Copy NTFS ACL's, defaults to ``True``
        :param str,optional cloud_folder: Target cloud folder, if ``create_cloud_folder_per_share`` was set to ``True`` then
         this attribute serves as the cloud folder name prefix
        :param bool,optional create_cloud_folder_per_share: Create cloud folder per share, defaults to ``False``
        :param bool,optional compute_checksum: Validate and report checksums post-migration, defaults to ``False``
        :param list(str),optional exclude: List of patterns to exclude, defaults to ``None``
        :param list(str),optional include: List of patterns to include, defaults to ``None``
        :param str,optional notes: Task notes

        :returns: Task
        :rtype: cterasdk.common.object.Object
        """
        param = TaskManager._create_add_parameter(name, credentials, shares, auto_start, notes)
        param.type = TaskType.Migration
        param.ntacl = int(winacls)
        param.calc_write_checksum = int(compute_checksum)
        param.cf = cloud_folder
        param.cf_per_share = int(create_cloud_folder_per_share)
        if access_time:
            param.atimes = int(access_time)
        if exclude:
            param.excludes = ':'.join(exclude)
        if include:
            param.excludes = ':'.join(include)
        return self._add(param)


class Task(Object):
    """Class representing a migration tool task"""

    def __init__(self, task_id, task_type, name, created_at=None, source=None, source_type=None, last_status=None, shares=None, notes=None):
        self.id = task_id
        self.type = {v: k for k, v in TaskType.__dict__.items() if not k.startswith('_')}.get(task_type).lower()
        self.name = name
        if created_at:
            self.created_at = created_at
        if source:
            self.source = source
        if source_type:
            sources = {v: k for k, v in SourceType.__dict__.items() if not k.startswith('_')}
            self.source_type = sources.get(source_type, None) if source_type else 'other'
        if last_status:
            self.last_status = last_status
        if shares:
            self.shares = [share.src for share in shares]
        if notes:
            self.notes = notes

    @staticmethod
    def from_server_object(server_object):
        parameters = dict(
            task_id=server_object.task_id,
            task_type=server_object.type,
            name=server_object.name,
            created_at=server_object.created_time,
            source=server_object.host,
            source_type=server_object.host_type,
            last_status=server_object.status_text.lower(),
            shares=server_object.shares,
            notes=server_object.notes
        )
        if server_object.type == TaskType.Discovery:
            parameters.update(dict(
                log_every_file=bool(server_object.discovery_log_files)
            ))
            return DiscoveryTask(**parameters)
        if server_object.type == TaskType.Migration:
            parameters.update(dict(
                winacls=bool(server_object.ntacl),
                cloud_folder=server_object.cf,
                create_cloud_folder_per_share=server_object.cf_per_share,
                compute_checksum=bool(server_object.calc_write_checksum),
                exclude=server_object.excludes,
                include=server_object.includes,
                access_time=getattr(server_object, 'atimes', None),
                schedule=f'{server_object.schedule_date.replace("/", "-")} {server_object.schedule_time}'
                         if server_object.schedule_date else None
            ))

            throttling_rule = getattr(server_object, 'bwlimit', None)
            throttling = None
            if throttling_rule:
                throttling = Object()
                throttling.kbps = throttling_rule.limit
                throttling.start = getattr(throttling_rule, 'from', None)
                throttling.end = getattr(throttling_rule, 'to', None)
            parameters['throttling'] = throttling

            return MigrationTask(**parameters)
        return None


class DiscoveryTask(Task):
    """Class representing a migration tool discovery task"""

    def __init__(self, task_id, task_type, name, created_at, source, source_type,  # pylint: disable=too-many-arguments
                 last_status, shares, notes, log_every_file):
        super().__init__(task_id, task_type, name, created_at, source, source_type, last_status, shares, notes)
        self.log_every_file = log_every_file


class MigrationTask(Task):
    """Class representing a migration tool migration task"""

    def __init__(self, task_id, task_type, name, created_at, source, source_type,  # pylint: disable=too-many-arguments, too-many-locals
                 last_status, shares, notes, winacls, cloud_folder, create_cloud_folder_per_share,
                 compute_checksum, exclude, include, access_time, schedule, throttling):
        super().__init__(task_id, task_type, name, created_at, source, source_type, last_status, shares, notes)
        self.winacls = winacls
        self.cloud_folder = cloud_folder
        self.create_cloud_folder_per_share = create_cloud_folder_per_share
        self.compute_checksum = compute_checksum
        self.exclude = exclude
        self.include = include
        self.access_time = access_time
        self.schedule = schedule
        self.throttling = throttling



import logging

from ..common import union, Object
from .base_command import BaseCommand
from .types import ScheduledTask, BackgroundTask
from . import query
from ..exceptions import CTERAException, ObjectNotFoundException


class Servers(BaseCommand):
    """
    Global Admin Servers APIs
    """

    default = ['name']

    def __init__(self, portal):
        super().__init__(portal)
        self.tasks = Tasks(self._core)

    def _get_entire_object(self, server):
        try:
            return self._core.api.get(f'/servers/{server}')
        except CTERAException as error:
            raise CTERAException('Failed to retreive server', error)

    def get(self, name, include=None):
        """
        Retrieve server properties

        :param str name: Name of the server
        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: The server, including the requested fields
        """
        include = union(include or [], Servers.default)
        include = ['/' + attr for attr in include]
        server = self._core.api.get_multi(f'/servers/{name}', include)
        if server.name is None:
            raise ObjectNotFoundException('Could not find server', f'/servers/{name}', name=name)
        return server

    def list_servers(self, include=None):
        """
        Retrieve the servers that comprise CTERA Portal.\n
        To retrieve servers, you must first browse the Global Administration Portal, using: `GlobalAdmin.portals.browse_global_admin()`

        :param list[str],optional include: List of fields to retrieve, defaults to ['name']
        """
        # browse administration
        include = union(include or [], Servers.default)
        param = query.QueryParamBuilder().include(include).build()
        return query.iterator(self._core, '/servers', param)

    def modify(self, name, server_name=None, app=None, preview=None, enable_public_ip=None, public_ip=None,
               allow_user_login=None, enable_replication=None, replica_of=None):
        """
        Modify a Portal server

        :param str name: The current server name
        :param str,optional server_name: New server name
        :param bool,optional app: Application server
        :param bool,optional preview: Preview server
        :param bool,optional enable_public_ip: Enable or disable public NAT address
        :param str,optional public_ip: Public NAT address
        :param bool,optional allow_user_login: Allow or disallow logins to this server
        :param bool,optional enable_replication: Enable or disable database replication
        :param str,optional replica_of: Configure as a replicate of another Portal server. `enable_replication` must be set to `True`
        """

        server = self._get_entire_object(name)
        if enable_replication is True and replica_of is not None:
            server.replicationSettings = Object()
            server.replicationSettings._classname = 'ServerReplicationSettings'  # pylint: disable=protected-access
            server.replicationSettings.replicationOf = self.get(replica_of, ['baseObjectRef']).baseObjectRef
        if enable_replication is False:
            server.replicationSettings = None
        if server_name is not None:
            server.name = server_name
        if app is not None:
            server.isApplicationServer = app
        if preview is not None:
            server.renderingServer = preview
        if enable_public_ip is True and public_ip is not None:
            server.publicIpaddr = public_ip
        elif enable_public_ip is False:
            server.publicIpaddr = None
        if allow_user_login is not None:
            server.allowUserLogin = allow_user_login

        try:
            response = self._core.api.put(f'/servers/{name}', server)
            logging.getLogger().info("Server modified. %s", {'server': name})
            return response
        except CTERAException as error:
            logging.getLogger().error("Could not modify server.")
            raise CTERAException('Could not modify server', error)


class Tasks(BaseCommand):

    def background(self, name):
        """
        Get all background tasks

        :param str name: Name of the server
        :return: List of tasks
        """
        return [BackgroundTask.from_server_object(task, f'servers/{name}/bgTasks/{task.id}')
                for task in self._core.api.get(f'/servers/{name}/bgTasks')]

    def scheduled(self, name):
        """
        Get all scheduled tasks

        :param str name: Name of the server
        :return: List of tasks
        """
        return [ScheduledTask.from_server_object(task) for task in self._core.api.get(f'/servers/{name}/schedTasks')]



import json
import queue
import logging
from xml.etree.ElementTree import fromstring, ParseError

from .types import XMLTypes
from ..common import Item, Object, Device


def ParseValue(data):
    if not data:
        return data

    try:
        if "." in data:
            return float(data)
    except ValueError:
        pass

    try:
        return int(data)
    except ValueError:
        pass

    text = data.strip()
    if text == "true":
        return True
    if text == "false":
        return False

    return text


def SetAppendValue(item, value):
    if item.parent is not None:
        if isinstance(item.parent.value, list):
            item.parent.value.append(value)
        else:
            setattr(item.parent.value, item.id, value)
    else:
        item.value = value


def fromjsonstr(fromstr):
    if not fromstr:
        return fromstr

    root = Item()
    try:
        root.node = json.loads(fromstr)
    except json.decoder.JSONDecodeError:
        logging.getLogger().debug('Parsing failed. Expected JSON format but received invalid input.')
        return None
    root.parent = None
    root.value = None

    q = queue.Queue()
    q.put(root)
    while not q.empty():
        item = q.get()
        if item.node is None:
            SetAppendValue(item, None)
        elif isinstance(item.node, (int, float, bool, str)):
            SetAppendValue(item, item.node)
        elif isinstance(item.node, list):
            item.value = []
            SetAppendValue(item, item.value)
            for kidnode in item.node:
                kid = Item()
                kid.parent = item
                kid.node = kidnode
                q.put(kid)
        elif isinstance(item.node, dict):
            item.value = Object()
            SetAppendValue(item, item.value)
            for kidnode, kidvalue in item.node.items():
                kid = Item()
                kid.parent = item
                kid.id = kidnode
                kid.node = kidvalue
                q.put(kid)

    return root.value


def fromxmlstr(string):  # pylint: disable=too-many-branches,too-many-statements

    if not string:
        logging.getLogger().debug('Empty payload received.')

    root = Item()
    root.value = None
    root.parent = None

    try:
        root.node = fromstring(string)
    except ParseError:
        logging.getLogger().debug('Parsing failed. Expected XML format but received invalid input.')
        return None

    q = queue.Queue()
    q.put(root)
    while not q.empty():
        item = q.get()
        if item.node.tag == XMLTypes.VAL:
            value = ParseValue(item.node.text)
            SetAppendValue(item, value)
        elif item.node.tag == XMLTypes.LIST:
            item.value = []
            SetAppendValue(item, item.value)
            for kidnode in item.node:
                if kidnode.tag in [XMLTypes.OBJ, XMLTypes.VAL]:
                    kid = Item()
                    kid.parent = item
                    kid.node = kidnode
                    q.put(kid)
        elif item.node.tag == XMLTypes.OBJ:
            classname = item.node.attrib.get(XMLTypes.CLASS)
            uuid = item.node.attrib.get(XMLTypes.UUID)

            item.value = Object()
            if classname is not None:  # Convert <obj class="ShareConfig"> to { "_classname" : "ShareConfig" }
                item.value._classname = classname  # pylint: disable=protected-access
            if uuid is not None:  # Convert <obj uuid="6f0e8c79-..."> to { "_uuid" : "6f0e8c79-..." }
                item.value._uuid = uuid  # pylint: disable=protected-access

            SetAppendValue(item, item.value)

            for kidnode in item.node:
                if kidnode.tag == XMLTypes.ATT:
                    kid = Item()
                    kid.id = kidnode.attrib[XMLTypes.ID]
                    kid.parent = item
                    kid.node = kidnode
                    q.put(kid)
        elif item.node.tag == XMLTypes.ATT:
            if len(item.node) > 0:
                for kidnode in item.node:
                    if kidnode.tag in [XMLTypes.OBJ, XMLTypes.LIST, XMLTypes.VAL]:
                        kid = Item()
                        kid.id = item.id
                        kid.parent = item.parent
                        kid.node = kidnode
                        q.put(kid)
            else:
                SetAppendValue(item, None)              # include empty attrs
        elif item.node.tag == XMLTypes.DB:              # db.xml
            item.value = Device(
                item.node.attrib.get(XMLTypes.ID),
                item.node.attrib.get(XMLTypes.VERSION),
                item.node.attrib.get(XMLTypes.FIRMWARE)
            )
            SetAppendValue(item, item.value)

            kid = Item()
            kid.id = item.node.attrib.get(XMLTypes.ID)
            kid.parent = item
            kid.node = list(item.node)[0]
            q.put(kid)

    return root.value



import logging

from .base_command import BaseCommand
from ...exceptions import CTERAException


class Login(BaseCommand):
    """
    CTERA Portal Login APIs
    """

    async def login(self, username, password):
        """
        Log into the portal

        :param str username: User name to log in
        :param str password: User password
        """
        host = self._core.host()
        try:
            await self._core.v1.api.form_data('/login', {'j_username': username, 'j_password': password})
            logging.getLogger().info("User logged in. %s", {'host': host, 'user': username})
        except CTERAException:
            logging.getLogger().error("Login failed. %s", {'host': host, 'user': username})
            raise

    async def logout(self):
        """
        Log out of the portal
        """
        await self._core.v1.api.form_data('/logout', {})
        logging.getLogger().info("User logged out. %s", {'host': self._core.host()})



import os
import sys
import logging


class Logging:  # pylint: disable=unused-private-member

    __instance = None

    @staticmethod
    def get():
        if Logging.__instance is None:
            Logging()
        return Logging.__instance

    def __init__(self):
        if Logging.__instance is not None:
            raise Exception("You must invoke Logging.get()")
        disabled = logconf['disabled']
        level = logconf['level']
        fmt = logconf['fmt']
        df = logconf['df']
        filename = logconf['filename']
        Logging.__instance = self
        self._logconf(fmt, df, filename=filename)
        logger = logging.getLogger()
        logger.disabled = disabled
        logger.level = level

    @staticmethod
    def _logconf(fmt, df, filename=None):
        basic_config_params = {
            'format': fmt,
            'datefmt': df,
        }
        if filename:
            basic_config_params['filename'] = filename
        else:
            basic_config_params['stream'] = sys.stdout

        logging.basicConfig(**basic_config_params)

    @staticmethod
    def disable():
        logging.getLogger().disabled = True

    @staticmethod
    def enable():
        logging.getLogger().disabled = False

    @staticmethod
    def setLevel(level):
        logging.getLogger().setLevel(level)

    @staticmethod
    def fmt():
        return logconf['fmt']

    @staticmethod
    def df():
        return logconf['df']


logconf = dict(
    disabled=False,
    level=logging.INFO,
    fmt='%(asctime)s,%(msecs)3d %(levelname)7s [%(filename)s:%(lineno)d] [%(funcName)s] - %(message)s',
    df='%Y-%m-%d %H:%M:%S',
    filename=os.environ.get('CTERASDK_LOG_FILE')
)




import time
import asyncio
import logging
import cterasdk.settings
from . import errors
from ..base import BaseClient
from ..common import Serializers
from ..asynchronous.clients import AsyncResponse
from ...common import Object
from .. import async_requests, decorators


class Client(BaseClient):
    """Synchronous Client"""

    def __init__(self, builder=None, async_session=None, authenticator=None):
        super().__init__(builder, async_session, authenticator, **session_settings())

    @decorators.authenticated
    def handle(self, path, *, on_response=None, **kwargs):
        request = async_requests.GetRequest(self._builder(path), **kwargs)
        return self._request(request, on_response=on_response)

    @decorators.authenticated
    def get(self, path, *, on_response=None, **kwargs):
        request = async_requests.GetRequest(self._builder(path), **kwargs)
        return self._request(request, on_response=on_response)

    @decorators.authenticated
    def put(self, path, data, *, data_serializer=None, on_response=None, **kwargs):
        request = async_requests.PutRequest(self._builder(path), data=data_serializer(data), **kwargs)
        return self._request(request, on_response=on_response)

    @decorators.authenticated
    def post(self, path, data, *, data_serializer=None, on_response=None, **kwargs):
        request = async_requests.PostRequest(self._builder(path), data=data_serializer(data), **kwargs)
        return self._request(request, on_response=on_response)

    @decorators.authenticated
    def form_data(self, path, data, *, on_response=None, **kwargs):
        request = async_requests.PostRequest(self._builder(path), data=Serializers.FormData(data), **kwargs)
        return self._request(request, on_response=on_response)

    @decorators.authenticated
    def multipart(self, path, form, *, on_response=None, **kwargs):
        request = async_requests.PostRequest(self._builder(path), data=form.data, **kwargs)
        return self._request(request, on_response=on_response)

    @decorators.authenticated
    def delete(self, path, *, on_response=None, **kwargs):
        request = async_requests.DeleteRequest(self._builder(path), **kwargs)
        return self._request(request, on_response=on_response)

    def _request(self, request, *, on_response=None):
        on_response = on_response if on_response else SyncResponse.new()
        return execute_request(self._async_session, self.join_headers(request), on_response=on_response)

    def shutdown(self):  # pylint: disable=invalid-overridden-method
        return asyncio.get_event_loop().run_until_complete(super().shutdown())


class Folders(Client):

    def download_zip(self, path, data, **kwargs):
        return super().form_data(path, data, **kwargs)


class Upload(Client):

    def upload(self, path, data, **kwargs):
        return super().form_data(path, data, **kwargs)


class Dav(Client):
    """WebDAV"""

    def download(self, path, **kwargs):
        return super().handle(path, **kwargs)

    def mkcol(self, path):
        request = async_requests.MkcolRequest(self._builder(path))
        response = self._request(request)
        return response.text()

    def copy(self, source, destination, *, overwrite=False):
        request = async_requests.CopyRequest(self._builder(source), headers=self._webdav_headers(destination, overwrite))
        response = self._request(request)
        return response.xml()

    def move(self, source, destination, *, overwrite=False):
        request = async_requests.MoveRequest(self._builder(source), headers=self._webdav_headers(destination, overwrite))
        response = self._request(request)
        return response.xml()

    def delete(self, path):  # pylint: disable=arguments-differ
        response = super().delete(path)
        return response.text()

    def _webdav_headers(self, destination, overwrite):
        return {
            'Destination': self._builder(destination),
            'Overwrite': 'T' if overwrite is True else 'F'
        }


class XML(Client):
    """XML Serializer and Deserializer"""

    def get(self, path, **kwargs):
        response = super().get(path, **kwargs)
        return response.xml()

    def put(self, path, data, **kwargs):
        response = super().put(path, data, data_serializer=Serializers.XML, **kwargs)
        return response.xml()

    def post(self, path, data, **kwargs):
        response = super().post(path, data, data_serializer=Serializers.XML, **kwargs)
        return response.xml()

    def form_data(self, path, data, **kwargs):
        response = super().form_data(path, data, **kwargs)
        return response.xml()

    def delete(self, path, **kwargs):
        response = super().delete(path, **kwargs)
        return response.xml()


class JSON(Client):
    """JSON Serializer and Deserializer"""

    def get(self, path, **kwargs):
        response = super().get(path, **kwargs)
        return response.json()

    def put(self, path, data, **kwargs):
        response = super().put(path, data, data_serializer=Serializers.JSON, **kwargs)
        return response.json()

    def post(self, path, data, **kwargs):
        response = super().post(path, data, data_serializer=Serializers.JSON, **kwargs)
        return response.json()

    def delete(self, path, **kwargs):
        response = super().delete(path, **kwargs)
        return response.json()

    def request(self, request):
        return super()._request(request)


class Extended(XML):
    """CTERA Schema"""

    def get_multi(self, path, paths):
        return self.database(path, 'get-multi', paths)

    def show_multi(self, path, paths):
        print(Serializers.JSON(self.get_multi(path, paths), no_log=False))

    def execute(self, path, name, param=None):  # schema method
        return self._execute(path, 'user-defined', name, param)

    def database(self, path, name, param=None):  # schema method
        return self._execute(path, 'db', name, param)

    def add(self, path, param=None):
        return self.database(path, 'add', param)

    def _execute(self, path, _type, name, param):
        data = Object()
        data.type = _type
        data.name = name
        data.param = param
        return super().post(path, data)


class API(Extended):
    """CTERA Management API"""

    def defaults(self, classname):
        return self.get(f'/defaults/{classname}')


class Migrate(JSON):
    """CTERA Migrate Service"""

    ID = 'x-mt-x'

    def login(self):
        request = async_requests.GetRequest(self._builder('/auth/user'))
        response = self.request(request)
        token = response.headers.get(Migrate.ID, None)
        if token:
            self.headers.update_headers({Migrate.ID: token})
        return response.json()


def execute_request(async_session, request, *, on_response, max_retries=3, backoff_factor=2):
    retries = 0
    while retries < max_retries:
        try:
            response = asyncio.get_event_loop().run_until_complete(async_session.await_promise(request, on_response=on_response))
            return errors.accept(response)
        except (ConnectionError, TimeoutError):
            retries += 1
            if retries < max_retries:
                delay = backoff_factor ** retries
                logging.getLogger().warning("Retrying in %s seconds.", delay)
                time.sleep(delay)
            else:
                logging.getLogger().error("Max retries reached. Request failed.")
                raise
    return None


class SyncResponse(AsyncResponse):
    """Synchronous Response Object"""

    def __init__(self, response):
        super().__init__(response)
        self._executor = asyncio.get_event_loop()

    def iter_content(self, chunk_size=None):
        while True:
            try:
                yield self._executor.run_until_complete(self.chunk(chunk_size).__anext__())
            except StopAsyncIteration:
                break

    def text(self):  # pylint: disable=invalid-overridden-method
        return self._consume_response(super().text)

    def json(self):  # pylint: disable=invalid-overridden-method
        return self._consume_response(super().json)

    def xml(self):  # pylint: disable=invalid-overridden-method
        return self._consume_response(super().xml)

    def _consume_response(self, consumer):
        return self._executor.run_until_complete(consumer())

    @staticmethod
    def new():
        async def new_response(response):
            return SyncResponse(response)
        return new_response


def session_settings():
    return {
        'connector': async_requests.create_connector(ssl=cterasdk.settings.sessions.management.ssl),
        'cookie_jar': async_requests.create_cookie_jar(unsafe=cterasdk.settings.sessions.management.allow_unsafe)
    }



import logging

from ..lib.task_manager_base import TaskError
from .enum import VolumeStatus
from ..common import Object
from ..exceptions import CTERAException, InputError
from ..lib import track
from .base_command import BaseCommand


class Volumes(BaseCommand):
    """ Edge Filer Volumes configuration APIs """

    def get(self, name=None):
        """
        Get Volume. If a volume name was not passed as an argument, a list of all storage volumes will be retrieved
        :param str,optional name: Name of the volume
        """
        return self._edge.api.get('/config/storage/volumes' + ('' if name is None else ('/' + name)))

    def add(self, name, size=None, filesystem='xfs', device=None, passphrase=None):
        """
        Add a Volume.

        :param str name: Name of the new volume
        :param int,optional size: Size of the new volume, defaults to the device's size
        :param str,optional filesystem: Filesystem to use, defaults to xfs
        :param str,optional device: Name of the device to use for the new volume, can be left as None if there's only one device available
        :param str,optional passphrase: Passphrase for the volume

        :return: Edge Filer response
        """
        storage_devices = self._devices()
        device_name, device_size = Volumes._device_volume(device, storage_devices)
        storage_volume_size = Volumes._volume_size(size, device_name, device_size)
        filesystem = Volumes._volume_filesystem(filesystem)

        param = Object()
        param.name = name
        param.device = device_name
        param.size = storage_volume_size
        param.fileSystemType = filesystem

        if passphrase is not None:
            param.encrypted = True
            param.encPassphrase = passphrase

        logging.getLogger().info(
            'Creating volume. %s',
            {'name': name, 'device': device_name, 'size': size, 'filesystem': filesystem, 'passphrase': passphrase}
        )

        ref = '/status/storage/volumes/' + param.name + '/status'

        response = self._edge.api.add('/config/storage/volumes', param)

        status = track(
            self._edge,
            ref,
            [VolumeStatus.Ok],
            [VolumeStatus.Formatting],
            [VolumeStatus.Mounting, VolumeStatus.Checking, VolumeStatus.Repairing], [VolumeStatus.Corrupted, VolumeStatus.Unknown]
        )

        logging.getLogger().info('Volume created. %s', {'name': name, 'size': size, 'status': status})

        return response

    def modify(self, name, size=None):
        """
        Modify an existing volume

        :param int,optional size: New size of the volume, if not set, the size will not change
        :return: Edge Filer response
        """
        if size is None:
            return Object()

        try:
            volume = self.get(name)
        except CTERAException as error:
            raise CTERAException('Failed to get the volume', error)

        volume.size = size
        try:
            response = self._edge.api.put('/config/storage/volumes/' + name, volume)
            logging.getLogger().info("Volume modified. %s", {'volume': volume.name})
            return response
        except CTERAException as error:
            msg = "Failed to modify volume."
            logging.getLogger().error(msg)
            raise CTERAException(msg, error)

    def delete(self, name):
        """
        Delete a volume

        :param str name: Name of the volume to delete
        """
        self._wait_pending_mount(name)
        try:
            logging.getLogger().info('Deleting volume. %s', {'name': name})

            response = self._edge.api.delete('/config/storage/volumes/' + name)

            logging.getLogger().info("Volume deleted. %s", {'name': name})

            return response
        except CTERAException as error:
            logging.getLogger().error("Volume deletion failed. %s", {'name': name})
            raise CTERAException('Volume deletion falied', error)

    def delete_all(self):
        """ Delete all volumes """
        self._wait_pending_filesystem_mounts()

        volumes = self._edge.api.get('/config/storage/volumes')
        volume_count = len(volumes)
        if volume_count > 0:
            for volume in volumes:
                self.delete(volume.name)
        else:
            logging.getLogger().info('No volumes found.')

    def _devices(self):
        arrays = self._edge.api.get('/status/storage/arrays')
        drives = self._edge.api.get('/status/storage/disks')

        ctera_devices = {}
        for array in arrays:
            ctera_devices[array.name] = array.availableCapacity

        for drive in drives:
            ctera_devices[drive.name] = drive.availableCapacity

        if len(ctera_devices) == 0:
            logging.getLogger().error('Could not find any drives or arrays.')

            raise CTERAException('Could not find any drives or arrays')

        return ctera_devices

    @staticmethod
    def _device_volume(device_name, ctera_devices):
        if device_name is not None:
            device_size = ctera_devices.get(device_name)
            if device_size is not None:
                logging.getLogger().debug('Found drive. %s', {'name': device_name, 'size': device_size})
                return (device_name, device_size)

            device_names = [k for k, v in ctera_devices.items()]
            logging.getLogger().error('Invalid device name. %s', {'name': device_name})
            raise InputError('Invalid device name', device_name, device_names)

        if len(ctera_devices) == 1:
            device_name, device_size = ctera_devices.popitem()
            logging.getLogger().debug('Found drive. %s', {'name': device_name, 'size': device_size})
            return device_name, device_size

        device_names = [k for k, v in ctera_devices.items()]
        logging.getLogger().error('You must specify a drive or an array name. %s', {'options': device_names})
        raise CTERAException('You must specify a drive or an array name', None, options=device_names)

    @staticmethod
    def _volume_size(size, device_name, device_size):
        if size is not None:
            if size > device_size:
                logging.getLogger().error('You cannot exceed the available storage capacity. %s', {'size': size, 'free_size': device_size})
                raise InputError("You cannot exceed the available storage capacity", size, device_size)
            return size

        if device_size > 0:
            logging.getLogger().warning('You did not specify a volume size.')
            logging.getLogger().warning('Allocating available storage capacity. %s', {'name': device_name, 'free_size': device_size})
            return device_size

        logging.getLogger().error('Insufficient storage space. %s', {'name': device_name})
        raise CTERAException('Insufficient storage space', None, device=device_name, free_size=device_size)

    @staticmethod
    def _volume_filesystem(filesystem):
        filesystem = filesystem.lower()
        if filesystem in ['xfs', 'next3', 'ext3']:
            return filesystem
        raise InputError('Invalid file system type', filesystem, ['xfs', 'next3', 'ext3'])

    def _wait_pending_filesystem_mounts(self):
        logging.getLogger().debug('Checking for pending mount tasks.')

        tasks = self._edge.tasks.running()
        for mount in tasks:
            if mount.name.startswith('Mounting'):
                self._wait_mount(mount.id)

    def _wait_pending_mount(self, volume):
        logging.getLogger().debug('Checking for pending mount tasks. %s', {'volume': volume})

        tasks = self._edge.tasks.by_name(' '.join(['Mounting', volume, 'file system']))
        for mount in tasks:
            self._wait_mount(mount.id)

    def _wait_mount(self, tid):
        try:
            self._edge.tasks.wait(tid)
        except TaskError:
            logging.getLogger().debug('Failed mounting volume. %s', {'tid': tid})
